"""
BondConstraints contains classes for all constraints related to bond length
between atoms.

.. inheritance-diagram:: fullrmc.Constraints.BondConstraints
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, raise_if_collected, reset_if_collected_out_of_date
from ..Core.Collection import get_caller_frames
from ..Core.Constraint import Constraint, SingularConstraint, RigidConstraint
from ..Core.bonds import full_bonds_coords




class BondConstraint(RigidConstraint, SingularConstraint):
    """
    Controls the bond's length defined between two atoms.

    +--------------------------------------------------------------------------+
    |.. figure:: bondSketch.png                                                |
    |   :width: 237px                                                          |
    |   :height: 200px                                                         |
    |   :align: center                                                         |
    |                                                                          |
    |   Bond sketch defined between two atoms.                                 |
    +--------------------------------------------------------------------------+

    .. raw:: html

        <iframe width="560" height="315"
        src="https://www.youtube.com/embed/GxmJae9h78E"
        frameborder="0" allowfullscreen>
        </iframe>


    +------------------------------------------------------------------------------+
    |.. figure:: bond_constraint_plot_method.png                                   |
    |   :width: 530px                                                              |
    |   :height: 400px                                                             |
    |   :align: left                                                               |
    +------------------------------------------------------------------------------+


    :Parameters:
        #. rejectProbability (Number): rejecting probability of all steps
           where standardError increases. It must be between 0 and 1 where 1
           means rejecting all steps where standardError increases and 0 means
           accepting all steps regardless whether standardError increases or not.

    .. code-block:: python

        ## Water (H2O) molecule sketch
        ##
        ##              O
        ##            /   \\
        ##         /   H2O   \\
        ##       H1           H2

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.BondConstraints import BondConstraint

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # create and add constraint
        BC = BondConstraint()
        ENGINE.add_constraints(BC)

        # define intra-molecular bonds
        BC.create_bonds_by_definition( bondsDefinition={"H2O": [ ('O','H1', 0.88, 1.02),
                                                                 ('O','H2', 0.88, 1.02) ]} )

    """

    def __init__(self, rejectProbability=1):
        # initialize constraint
        RigidConstraint.__init__(self, rejectProbability=rejectProbability)
        # set atomsCollector datakeys
        self._atomsCollector.set_data_keys( ['map',] )
        # init bonds data
        self.__bondsList = [[],[],[],[]]
        self.__bonds     = {}
        # set computation cost
        self.set_computation_cost(1.0)
        # create dump flag
        self.__dumpBonds = True
         # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_BondConstraint__bondsList',
                           '_BondConstraint__bonds',] )
        RUNTIME_DATA = [d for d in self.RUNTIME_DATA]
        RUNTIME_DATA.extend( [] )
        object.__setattr__(self, 'FRAME_DATA',   tuple(FRAME_DATA)   )
        object.__setattr__(self, 'RUNTIME_DATA', tuple(RUNTIME_DATA) )

    @property
    def bondsList(self):
        """ List of defined bonds"""
        return self.__bondsList

    @property
    def bonds(self):
        """ Bonds dictionary map of every and each atom"""
        return self.__bonds

    def _on_collector_reset(self):
        self._atomsCollector._randomData = set([])

    def listen(self, message, argument=None):
        """
        Listens to any message sent from the Broadcaster.

        :Parameters:
            #. message (object): Any python object to send to constraint's
               listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine set","update boundary conditions",):
            # set bonds and reset constraint
            self.set_bonds(self.__bondsList, tform=False)
            # reset constraint is called in set_bonds

    #@raise_if_collected
    def set_bonds(self, bondsList, tform=True):
        """
        Sets bonds dictionary by parsing bondsList list.

        :Parameters:
            #. bondsList (None, list): Bonds definition list. If None is given
               no bonds are defined. Otherwise it can be of any of the
               following two forms:

                   tuples format: every item must be a list or tuple of four
                   items.\n
                   #. First atom index forming the bond.
                   #. Second atom index forming the bond.
                   #. Lower limit or the minimum bond length
                      allowed.
                   #. Upper limit or the maximum bond length
                      allowed.

                   four vectors format: List of exaclty four lists or
                   numpy.arrays of the same length.\n
                   #. List contains the first atoms index
                      forming the bond.
                   #. List contains the second atoms index
                      forming the bond.
                   #. List containing the lower limit or
                      the minimum bond length allowed.
                   #. List containing the upper limit or
                      the maximum bond length allowed.

            #. tform (boolean): set whether given bondsList follows tuples
               format, If not then it must follow the four vectors one.
        """
        # check if bondsList is given
        if bondsList is None:
            bondsList = [[],[],[],[]]
            tform     = False
        elif len(bondsList) == 4 and len(bondsList[0]) == 0:
            tform     = False
        if self.engine is None:
            self.__bondsList = bondsList
            self.__bonds     = {}
        else:
            NUMBER_OF_ATOMS = self.engine.get_original_data("numberOfAtoms")
            # set bonds and bondsList definition
            oldBondsList = self.__bondsList
            oldBonds     = self.__bonds
            self.__bondsList = [np.array([], dtype=INT_TYPE),
                                np.array([], dtype=INT_TYPE),
                                np.array([], dtype=FLOAT_TYPE),
                                np.array([], dtype=FLOAT_TYPE)]
            self.__bonds     = {}
            self.__dumpBonds = False
            # build bonds
            try:
                if tform:
                    for bond in bondsList:
                        self.add_bond(bond)
                else:
                    for idx in xrange(len(bondsList[0])):
                        bond = [bondsList[0][idx], bondsList[1][idx], bondsList[2][idx], bondsList[3][idx]]
                        self.add_bond(bond)
            except Exception as e:
                self.__dumpBonds = True
                self.__bondsList = oldBondsList
                self.__bonds     = oldBonds
                LOGGER.error(e)
                import traceback
                raise Exception(traceback.format_exc())
            self.__dumpBonds = True
            # finalize bonds
            for idx in xrange(NUMBER_OF_ATOMS):
                self.__bonds[INT_TYPE(idx)] = self.__bonds.get(INT_TYPE(idx), {"indexes":[],"map":[]} )
        # dump to repository
        self._dump_to_repository({'_BondConstraint__bondsList':self.__bondsList,
                                  '_BondConstraint__bonds'    :self.__bonds})
        # reset constraint
        self.reset_constraint()

    #@raise_if_collected
    def add_bond(self, bond):
        """
        Add a single bond to constraint's list of bonds.

        :Parameters:
            #. bond (list): The bond list of four items.\n
               #. First atom index forming the bond.
               #. Second atom index forming the bond.
               #. Lower limit or the minimum bond length allowed.
               #. Upper limit or the maximum bond length allowed.
        """
        assert self.engine is not None, LOGGER.error("setting a bond is not allowed unless engine is defined.")
        NUMBER_OF_ATOMS = self.engine.get_original_data("numberOfAtoms")
        assert isinstance(bond, (list, set, tuple)), LOGGER.error("bond items must be lists")
        assert len(bond)==4, LOGGER.error("bond items must be lists of 4 items each")
        idx1, idx2, lower, upper = bond
        assert is_integer(idx1), LOGGER.error("bondsList items lists first item must be an integer")
        idx1 = INT_TYPE(idx1)
        assert is_integer(idx2), LOGGER.error("bondsList items lists second item must be an integer")
        idx2 = INT_TYPE(idx2)
        assert idx1<NUMBER_OF_ATOMS, LOGGER.error("bond atom index must be smaller than maximum number of atoms")
        assert idx2<NUMBER_OF_ATOMS, LOGGER.error("bond atom index must be smaller than maximum number of atoms")
        assert idx1>=0, LOGGER.error("bond first item must be positive")
        assert idx2>=0, LOGGER.error("bond second item must be positive")
        assert idx1!=idx2, LOGGER.error("bond first and second items can't be the same")
        assert is_number(lower), LOGGER.error("bond third item must be a number")
        lower = FLOAT_TYPE(lower)
        assert is_number(upper), LOGGER.error("bond fourth item must be a number")
        upper = FLOAT_TYPE(upper)
        assert lower>=0, LOGGER.error("bond third item must be positive")
        assert upper>lower, LOGGER.error("bond third item must be smaller than the fourth item")
        # create bonds
        if not idx1 in self.__bonds:
            bondsIdx1 = {"indexes":[],"map":[]}
        else:
            bondsIdx1 = {"indexes":self.__bonds[idx1]["indexes"],
                         "map"    :self.__bonds[idx1]["map"] }
        if not INT_TYPE(idx2) in self.__bonds:
            bondsIdx2 = {"indexes":[],"map":[]}
        else:
            bondsIdx2 = {"indexes":self.__bonds[idx2]["indexes"],
                         "map"    :self.__bonds[idx2]["map"] }
        # set bond
        if idx2 in bondsIdx1["indexes"]:
            assert idx1 in bondsIdx2["indexes"], LOOGER.error("mismatched bonds between atom '%s' and '%s'"%(idx1,idx2))
            at2InAt1 = bondsIdx1["indexes"].index(idx2)
            at1InAt2 = bondsIdx2["indexes"].index(idx1)
            assert bondsIdx1["map"][at2InAt1] == bondsIdx2["map"][at1InAt2], LOOGER.error("bonded atoms '%s' and '%s' point to different defintions"%(idx1,idx2))
            setPos = bondsIdx1["map"][at2InAt1]
            LOGGER.warn("Bond between atom index '%i' and '%i' is already defined. New bond limits [%.3f,%.3f] will replace old bond limits [%.3f,%.3f]. "%(idx2, idx1, lower, upper, self.__bondsList[2][setPos], self.__bondsList[3][setPos]))
            self.__bondsList[0][setPos] = idx1
            self.__bondsList[1][setPos] = idx2
            self.__bondsList[2][setPos] = lower
            self.__bondsList[3][setPos] = upper
        else:
            bondsIdx1["map"].append( len(self.__bondsList[0]) )
            bondsIdx2["map"].append( len(self.__bondsList[0]) )
            bondsIdx1["indexes"].append(idx2)
            bondsIdx2["indexes"].append(idx1)
            self.__bondsList[0] = np.append(self.__bondsList[0],idx1)
            self.__bondsList[1] = np.append(self.__bondsList[1],idx2)
            self.__bondsList[2] = np.append(self.__bondsList[2],lower)
            self.__bondsList[3] = np.append(self.__bondsList[3],upper)
        self.__bonds[idx1] = bondsIdx1
        self.__bonds[idx2] = bondsIdx2
        # dump to repository
        if self.__dumpBonds:
            self._dump_to_repository({'_BondConstraint__bondsList' :self.__bondsList,
                                      '_BondConstraint__bonds'     :self.__bonds})
            # reset constraint
            self.reset_constraint()

    #@raise_if_collected
    def create_bonds_by_definition(self, bondsDefinition):
        """
        Creates bondsList using bonds definition.
        Calls set_bonds(bondsList) and generates bonds attribute.

        :Parameters:
            #. bondsDefinition (dict): The bonds definition.
               Every key must be a molecule's name.
               Every key value must be a list of bonds definitions.
               Every bond definition is a list of four items where:

               #. Name of the first atom forming the bond.
               #. Name of the second atom forming the bond.
               #. Lower limit or the minimum bond length allowed.
               #. Upper limit or the maximum bond length allowed.

           ::

                e.g. (Carbon tetrachloride):  bondsDefinition={"CCL4": [('C','CL1' , 1.55, 1.95),
                                                                        ('C','CL2' , 1.55, 1.95),
                                                                        ('C','CL3' , 1.55, 1.95),
                                                                        ('C','CL4' , 1.55, 1.95) ] }

        """
        if self.engine is None:
            raise Exception(LOGGER.error("engine is not defined. Can't create bonds by definition"))
            return
        if bondsDefinition is None:
            bondsDefinition = {}
        assert isinstance(bondsDefinition, dict), LOGGER.error("bondsDefinition must be a dictionary")
        ALL_NAMES       = self.engine.get_original_data("allNames")
        NUMBER_OF_ATOMS = self.engine.get_original_data("numberOfAtoms")
        MOLECULES_NAME  = self.engine.get_original_data("moleculesName")
        MOLECULES_INDEX = self.engine.get_original_data("moleculesIndex")
        # check map definition
        existingmoleculesName = sorted(set(MOLECULES_NAME))
        bondsDef = {}
        for mol in bondsDefinition:
            bonds = bondsDefinition[mol]
            if mol not in existingmoleculesName:
                LOGGER.warn("Molecule name '%s' in bondsDefinition is not recognized, bonds definition for this particular molecule is omitted"%str(mol))
                continue
            assert isinstance(bonds, (list, set, tuple)), LOGGER.error("mapDefinition molecule bonds must be a list")
            bonds = list(bonds)
            molbondsList = []
            for bond in bonds:
                assert isinstance(bond, (list, set, tuple)), LOGGER.error("mapDefinition bonds must be a list")
                bond = list(bond)
                assert len(bond)==4, LOGGER.error("mapDefinition bonds list must be of length 4")
                at1, at2, lower, upper = bond
                assert is_number(lower), LOGGER.error("mapDefinition bonds list third item must be a number")
                lower = FLOAT_TYPE(lower)
                assert is_number(upper), LOGGER.error("mapDefinition bonds list fourth item must be a number")
                upper = FLOAT_TYPE(upper)
                assert lower>=0, LOGGER.error("mapDefinition bonds list third item must be bigger than 0")
                assert upper>lower, LOGGER.error("mapDefinition bonds list fourth item must be bigger than the third item")
                # check for redundancy
                append = True
                for b in molbondsList:
                    if (b[0]==at1 and b[1]==at2) or (b[1]==at1 and b[0]==at2):
                        LOGGER.warn("Redundant definition for bondsDefinition found. The later '%s' is ignored"%str(b))
                        append = False
                        break
                if append:
                    molbondsList.append((at1, at2, lower, upper))
            # create bondDef for molecule mol
            bondsDef[mol] = molbondsList
        # create mols dictionary
        mols = {}
        for idx in xrange(NUMBER_OF_ATOMS):
            molName = MOLECULES_NAME[idx]
            if not molName in bondsDef:
                continue
            molIdx = MOLECULES_INDEX[idx]
            if not molIdx in mols:
                mols[molIdx] = {"name":molName, "indexes":[], "names":[]}
            mols[molIdx]["indexes"].append(idx)
            mols[molIdx]["names"].append(ALL_NAMES[idx])
        # get bondsList
        bondsList = []
        for val in mols.values():
            indexes = val["indexes"]
            names   = val["names"]
            # get definition for this molecule
            thisDef = bondsDef[val["name"]]
            for bond in thisDef:
                idx1  = indexes[ names.index(bond[0]) ]
                idx2  = indexes[ names.index(bond[1]) ]
                lower = bond[2]
                upper = bond[3]
                bondsList.append((idx1, idx2, lower, upper))
        # create bonds
        self.set_bonds(bondsList=bondsList)

    def compute_standard_error(self, data):
        """
        Compute the standard error (StdErr) of data not satisfying constraint
        conditions.

        .. math::
            StdErr = \\sum \\limits_{i}^{C}
            ( \\beta_{i} - \\beta_{i}^{min} ) ^{2}
            \\int_{0}^{\\beta_{i}^{min}} \\delta(\\beta-\\beta_{i}) d \\beta
            +
            ( \\beta_{i} - \\beta_{i}^{max} ) ^{2}
            \\int_{\\beta_{i}^{max}}^{\\infty} \\delta(\\beta-\\beta_{i}) d \\beta

        Where:\n
        :math:`C` is the total number of defined bonds constraints. \n
        :math:`\\beta_{i}^{min}` is the bond constraint lower limit set for constraint i. \n
        :math:`\\beta_{i}^{max}` is the bond constraint upper limit set for constraint i. \n
        :math:`\\beta_{i}` is the bond length computed for constraint i. \n
        :math:`\\delta` is the Dirac delta function. \n
        :math:`\\int_{0}^{\\beta_{i}^{min}} \\delta(\\beta-\\beta_{i}) d \\beta`
        is equal to 1 if :math:`0 \\leqslant \\beta_{i} \\leqslant \\beta_{i}^{min}` and 0 elsewhere.\n
        :math:`\\int_{\\beta_{i}^{max}}^{\\pi} \\delta(\\beta-\\beta_{i}) d \\beta`
        is equal to 1 if :math:`\\beta_{i}^{max} \\leqslant \\beta_{i} \\leqslant \\infty` and 0 elsewhere.\n

        :Parameters:
            #. data (object): Data to compute standardError.

        :Returns:
            #. standardError (number): The calculated standardError of the
            given.
        """
        return FLOAT_TYPE( np.sum(data["reducedLengths"]**2) )

    def get_constraint_value(self):
        """
        Get constraint's data.

        :Returns:
            #. data (numpy.array): The constraint value data
        """
        return self.data

    @reset_if_collected_out_of_date
    def compute_data(self):
        """ Compute constraint's data."""
        if len(self._atomsCollector):
            bondsData   = np.zeros(self.__bondsList[0].shape[0], dtype=FLOAT_TYPE)
            reducedData = np.zeros(self.__bondsList[0].shape[0], dtype=FLOAT_TYPE)
            bondsIndexes = set(set(range(self.__bondsList[0].shape[0])))
            bondsIndexes = list( bondsIndexes-self._atomsCollector._randomData )
            idx1 = self._atomsCollector.get_relative_indexes(self.__bondsList[0][bondsIndexes])
            idx2 = self._atomsCollector.get_relative_indexes(self.__bondsList[1][bondsIndexes])
            lowerLimit = self.__bondsList[2][bondsIndexes]
            upperLimit = self.__bondsList[3][bondsIndexes]
        else:
            idx1 = self._atomsCollector.get_relative_indexes(self.__bondsList[0])
            idx2 = self._atomsCollector.get_relative_indexes(self.__bondsList[1])
            lowerLimit = self.__bondsList[2]
            upperLimit = self.__bondsList[3]
        # compute
        bonds, reduced = full_bonds_coords(idx1                  = idx1,
                                           idx2                  = idx2,
                                           lowerLimit            = lowerLimit,
                                           upperLimit            = upperLimit,
                                           boxCoords             = self.engine.boxCoordinates,
                                           basis                 = self.engine.basisVectors,
                                           isPBC                 = self.engine.isPBC,
                                           reduceDistanceToUpper = False,
                                           reduceDistanceToLower = False,
                                           ncores                = INT_TYPE(1))
        # create full length data
        if len(self._atomsCollector):
            bondsData[bondsIndexes]   = bonds
            reducedData[bondsIndexes] = reduced
            bonds   = bondsData
            reduced = reducedData
        self.set_data( {"bondsLength":bonds, "reducedLengths":reduced} )
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        self.set_standard_error( self.compute_standard_error(data = self.data) )
        # set original data
        if self.originalData is None:
            self._set_original_data(self.data)

    def compute_before_move(self, realIndexes, relativeIndexes):
        """
        Compute constraint's data before move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Group atoms index the move will
               be applied to.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # get bonds indexes
        bondsIndexes = []
        #for idx in relativeIndexes:
        for idx in realIndexes:
            bondsIndexes.extend( self.__bonds[idx]['map'] )
        #bondsIndexes = list(set(bondsIndexes))
        # remove collected bonds
        bondsIndexes = list( set(bondsIndexes)-set(self._atomsCollector._randomData) )
        # compute data before move
        if len(bondsIndexes):
            bonds, reduced = full_bonds_coords(idx1                  = self._atomsCollector.get_relative_indexes(self.__bondsList[0][bondsIndexes]),
                                               idx2                  = self._atomsCollector.get_relative_indexes(self.__bondsList[1][bondsIndexes]),
                                               lowerLimit            = self.__bondsList[2][bondsIndexes],
                                               upperLimit            = self.__bondsList[3][bondsIndexes],
                                               boxCoords             = self.engine.boxCoordinates,
                                               basis                 = self.engine.basisVectors,
                                               isPBC                 = self.engine.isPBC,
                                               reduceDistanceToUpper = False,
                                               reduceDistanceToLower = False,
                                               ncores                = INT_TYPE(1))
        else:
            bonds   = None
            reduced = None
        # set data before move
        self.set_active_atoms_data_before_move( {"bondsIndexes":bondsIndexes, "bondsLength":bonds, "reducedLengths":reduced} )
        self.set_active_atoms_data_after_move(None)

    def compute_after_move(self, realIndexes, relativeIndexes, movedBoxCoordinates):
        """
        Compute constraint after move is executed

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new
               coordinates.
        """
        bondsIndexes = self.activeAtomsDataBeforeMove["bondsIndexes"]
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[relativeIndexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[relativeIndexes] = movedBoxCoordinates
        # compute data after move
        if len(bondsIndexes):
            bonds, reduced = full_bonds_coords(idx1                  = self._atomsCollector.get_relative_indexes(self.__bondsList[0][bondsIndexes]),#self.__bondsList[0][bondsIndexes],
                                               idx2                  = self._atomsCollector.get_relative_indexes(self.__bondsList[1][bondsIndexes]),#self.__bondsList[1][bondsIndexes],
                                               lowerLimit            = self.__bondsList[2][bondsIndexes],
                                               upperLimit            = self.__bondsList[3][bondsIndexes],
                                               boxCoords             = self.engine.boxCoordinates,
                                               basis                 = self.engine.basisVectors,
                                               isPBC                 = self.engine.isPBC,
                                               reduceDistanceToUpper = False,
                                               reduceDistanceToLower = False,
                                               ncores                = INT_TYPE(1))
        else:
            bonds   = None
            reduced = None
        # set active data after move
        self.set_active_atoms_data_after_move( {"bondsLength":bonds, "reducedLengths":reduced} )
        # reset coordinates
        self.engine.boxCoordinates[relativeIndexes] = boxData
        # compute standardError after move
        if bonds is None:
            self.set_after_move_standard_error( self.standardError )
        else:
            # bondsIndexes is a fancy slicing, RL is a copy not a view.
            RL = self.data["reducedLengths"][bondsIndexes]
            self.data["reducedLengths"][bondsIndexes] += reduced-self.activeAtomsDataBeforeMove["reducedLengths"]
            self.set_after_move_standard_error( self.compute_standard_error(data = self.data) )
            self.data["reducedLengths"][bondsIndexes] = RL

    def accept_move(self, realIndexes, relativeIndexes):
        """
        Accept move

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # get bonds indexes
        bondsIndexes = self.activeAtomsDataBeforeMove["bondsIndexes"]
        if len(bondsIndexes):
            # set new data
            data = self.data
            data["bondsLength"][bondsIndexes]    += self.activeAtomsDataAfterMove["bondsLength"]-self.activeAtomsDataBeforeMove["bondsLength"]
            data["reducedLengths"][bondsIndexes] += self.activeAtomsDataAfterMove["reducedLengths"]-self.activeAtomsDataBeforeMove["reducedLengths"]
            self.set_data( data )
            # update standardError
            self.set_standard_error( self.afterMoveStandardError )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)

    def reject_move(self, realIndexes, relativeIndexes):
        """
        Reject move

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)

    def compute_as_if_amputated(self, realIndex, relativeIndex):
        """
        Compute and return constraint's data and standard error as if atom given its
        its was amputated.

        :Parameters:
            #. realIndex (numpy.ndarray): Not used here.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        pass

    def accept_amputation(self, realIndex, relativeIndex):
        """
        Accept amputation of atom and sets constraints data and standard error accordingly.

        :Parameters:
            #. realIndex (numpy.ndarray): Atom's index as a numpy array
               of a single element.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        # MAYBE WE DON"T NEED TO CHANGE DATA AND SE. BECAUSE THIS MIGHT BE A PROBLEM
        # WHEN IMPLEMENTING ATOMS RELEASING. MAYBE WE NEED TO COLLECT DATA INSTEAD, REMOVE
        # AND ADD UPON RELEASE
        # get all involved data
        bondsIndexes = []
        for idx in realIndex:
            bondsIndexes.extend( self.__bonds[idx]['map'] )
        bondsIndexes = list(set(bondsIndexes))
        if len(bondsIndexes):
            # set new data
            data = self.data
            data["bondsLength"][bondsIndexes]    = 0
            data["reducedLengths"][bondsIndexes] = 0
            self.set_data( data )
            # update standardError
            SE = self.compute_standard_error(data = self.get_constraint_value())
            self.set_standard_error( SE )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)

    def reject_amputation(self, realIndex, relativeIndex):
        """
        Reject amputation of atom.

        :Parameters:
            #. realIndex (numpy.ndarray): Not used here.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        pass

    def _on_collector_collect_atom(self, realIndex):
        # get bond indexes
        BI = self.__bonds[realIndex]['map']
        # append all mapped bonds to collector's random data
        self._atomsCollector._randomData = self._atomsCollector._randomData.union( set(BI) )
        #print(realIndex, BI, self._atomsCollector._randomData)
        # collect atom bondIndexes
        self._atomsCollector.collect(realIndex, dataDict={'map':BI})

    def plot(self, ax=None, nbins=25, subplots=True, split=None,
                   wspace=0.3, hspace=0.3,
                   histTtype='bar', lineWidth=None, lineColor=None,
                   xlabel=True, xlabelSize=16,
                   ylabel=True, ylabelSize=16,
                   legend=True, legendCols=1, legendLoc='best',
                   title=True, titleStdErr=True, titleAtRem=True,
                   titleUsedFrame=True, show=True):
        """
        Plot bonds constraint's distribution histogram.

        :Parameters:
            #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
               If ax is given,  subplots parameters will be omitted.
               If None is given a new plot figure will be created.
            #. nbins (int): number of bins in histogram.
            #. subplots (boolean): Whether to add plot constraint on multiple
               axes.
            #. split (None, 'name', 'element'): To split plots into histogram
               per atom names, elements in addition to lower and upper bounds.
               If None is given, histograms will be built from lower and upper
               bounds only.
            #. wspace (float): The amount of width reserved for blank space
               between subplots, expressed as a fraction of the average axis
               width.
            #. hspace (float): The amount of height reserved for white space
               between subplots, expressed as a fraction of the average axis
               height.
            #. histTtype (string): the histogram type. optional among
               ['bar', 'barstacked', 'step', 'stepfilled']
            #. lineWidth (None, integer): bars contour line width. If None is
               given then default value will be given.
            #. lineColor (None, integer): bars contour line color. If None is
               given then default value will be given.
            #. xlabel (boolean): Whether to create x label.
            #. xlabelSize (number): The x label font size.
            #. ylabel (boolean): Whether to create y label.
            #. ylabelSize (number): The y label font size.
            #. legend (boolean): Whether to create the legend or not
            #. legendCols (integer): Legend number of columns.
            #. legendLoc (string): The legend location. Anything among
               'right', 'center left', 'upper right', 'lower right', 'best',
               'center', 'lower left', 'center right', 'upper left',
               'upper center', 'lower center' is accepted.
            #. title (boolean): Whether to create the title or not.
            #. titleStdErr (boolean): Whether to show constraint standard error
               value in title.
            #. titleAtRem (boolean): Whether to show engine's number of removed
               atoms.
            #. titleUsedFrame(boolean): Whether to show used frame name in
               title.
            #. show (boolean): Whether to render and show figure before
               returning.

        :Returns:
            #. figure (matplotlib Figure): matplotlib used figure.
            #. axes (matplotlib Axes, List): matplotlib axes or a list of axes.
        """
        def _get_bins(dmin, dmax, boundaries, nbins):
            # create bins
            delta = float(dmax-dmin)/float(nbins-1)
            bins  = range(nbins)
            bins  = [b*delta for b in bins]
            bins  = [b+dmin for b in bins]
            # check boundaries
            bidx = 0
            for b in sorted(boundaries):
                for i in range(bidx, len(bins)-1):
                    bidx = i
                    # exact match with boundary
                    if b==bins[bidx]:
                        break
                    # boundary between two bins, move closest bin to boundary
                    if bins[bidx] < b < bins[bidx+1]:
                        if b-bins[bidx] > bins[bidx+1]-b:
                            bins[bidx+1] = b
                        else:
                            bins[bidx]   = b
                        break
            # return bins
            return bins
        # get constraint value
        output = self.get_constraint_value()
        if output is None:
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # import matplotlib
        import matplotlib.pyplot as plt
        # compute categories
        if split == 'name':
            splitV = self.engine.get_original_data("allNames")
        elif split == 'element':
            splitV = self.engine.get_original_data("allElements")
        else:
            splitV = None
        atom1 = self.__bondsList[0]
        atom2 = self.__bondsList[1]
        lower = self.__bondsList[2]
        upper = self.__bondsList[3]
        categories = {}
        for idx in xrange(self.__bondsList[0].shape[0]):
            if self._atomsCollector.is_collected(idx):
                continue
            if splitV is not None:
                a1 = splitV[ atom1[idx] ]
                a2 = splitV[ atom2[idx] ]
            else:
                a1 = a2 = ''
            l  = lower[idx]
            u  = upper[idx]
            k  = (a1,a2,l,u)
            L  = categories.get(k, [])
            L.append(idx)
            categories[k] = L
        ncategories = len(categories)
        # get axes
        if ax is None:
            if subplots and ncategories>1:
                x = np.ceil(np.sqrt(ncategories))
                y = np.ceil(ncategories/x)
                FIG, N_AXES = plt.subplots(int(x), int(y) )
                N_AXES = N_AXES.flatten()
                FIG.subplots_adjust(wspace=wspace, hspace=hspace)
                [N_AXES[i].axis('off') for i in range(ncategories,len(N_AXES))]
            else:
                FIG  = plt.figure()
                AXES = FIG.gca()
                subplots = False
        else:
            AXES = ax
            FIG = AXES.get_figure()
            subplots = False
        # start plotting
        COLORS = ["b",'g','r','c','y','m']
        if subplots:
            for idx, key in enumerate(categories):
                a1,a2, L,U  = key
                # get label
                label = "%s%s%s(%.2f,%.2f)"%(a1,'-'*(len(a1)>0),a2,L,U)
                COL   = COLORS[idx%len(COLORS)]
                AXES  = N_AXES[idx]
                idxs  = categories[key]
                data  = self.data["bondsLength"][idxs]
                # get data limits
                mn = np.min(data)
                mx = np.max(data)
                # get bins
                BINS = _get_bins(dmin=mn, dmax=mx, boundaries=[L,U], nbins=nbins)
                # plot histogram
                D, _, P = AXES.hist(x=data, bins=BINS,
                                    color=COL, label=label,
                                    histtype=histTtype)
                # vertical lines
                Y = max(D)
                AXES.plot([L,L],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle='--')
                AXES.plot([U,U],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle='--')
                # legend
                if legend:
                    AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
                # set axis labels
                if xlabel:
                    AXES.set_xlabel("$r(\\AA)$", size=xlabelSize)
                if ylabel:
                    AXES.set_ylabel("$number$"  , size=ylabelSize)
                if lineWidth is not None:
                    [p.set_linewidth(lineWidth) for p in P]
                if lineColor is not None:
                    [p.set_edgecolor(lineColor) for p in P]
                # update limits
                AXES.set_xmargin(0.1)
                AXES.autoscale()
        else:
            for idx, key in enumerate(categories):
                a1,a2, L,U  = key
                label = "%s%s%s(%.2f,%.2f)"%(a1,'-'*(len(a1)>0),a2,L,U)
                COL   = COLORS[idx%len(COLORS)]
                idxs  = categories[key]
                data  = self.data["bondsLength"][idxs]
                # get data limits
                mn = np.min(data)
                mx = np.max(data)
                # get bins
                BINS = _get_bins(dmin=mn, dmax=mx, boundaries=[L,U], nbins=nbins)
                # plot histogram
                D, _, P = AXES.hist(x=data, bins=BINS,
                                    color=COL, label=label,
                                    histtype=histTtype)
                # vertical lines
                Y = max(D)
                AXES.plot([L,L],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle='--')
                AXES.plot([U,U],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle='--')
                if lineWidth is not None:
                    [p.set_linewidth(lineWidth) for p in P]
                if lineColor is not None:
                    [p.set_edgecolor(lineColor) for p in P]
            # legend
            if legend:
                AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
            # set axis labels
            if xlabel:
                AXES.set_xlabel("$r(\\AA)$", size=xlabelSize)
            if ylabel:
                AXES.set_ylabel("$number$"  , size=ylabelSize)
            # update limits
            AXES.set_xmargin(0.1)
            AXES.autoscale()
        # set title
        if title:
            FIG.canvas.set_window_title('Bond Constraint')
            if titleUsedFrame:
                t = '$frame: %s$ : '%self.engine.usedFrame.replace('_','\\_')
            else:
                t = ''
            if titleAtRem:
                t += "$%i$ $rem.$ $at.$ - "%(len(self.engine._atomsCollector))
            if titleStdErr and self.standardError is not None:
                t += "$std$ $error=%.6f$ "%(self.standardError)
            if len(t):
                FIG.suptitle(t, fontsize=14)
        # set background color
        FIG.patch.set_facecolor('white')
        #show
        if show:
            plt.show()
        # return axes
        if subplots:
            return FIG, N_AXES
        else:
            return FIG, AXES

    def export(self, fname, delimiter='     ', comments='# ', split=None):
        """
        Export bonds constraint's distribution histogram.

        :Parameters:
            #. fname (path): full file name and path.
            #. delimiter (string): String or character separating columns.
            #. comments (string): String that will be prepended to the header.
            #. split (None, 'name', 'element'): To split output into per atom
               names, elements in addition to lower and upper bounds.
               If None is given, output will be built from lower and upper
               bounds only.
        """
        # get constraint value
        output = self.get_constraint_value()
        if output is None:
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # compute categories
        if split == 'name':
            splitV = self.engine.get_original_data("allNames")
        elif split == 'element':
            splitV = self.engine.get_original_data("allElements")
        else:
            splitV = None
        atom1 = self.__bondsList[0]
        atom2 = self.__bondsList[1]
        lower = self.__bondsList[2]
        upper = self.__bondsList[3]
        categories = {}
        for idx in xrange(self.__bondsList[0].shape[0]):
            if self._atomsCollector.is_collected(idx):
                continue
            if splitV is not None:
                a1 = splitV[ atom1[idx] ]
                a2 = splitV[ atom2[idx] ]
            else:
                a1 = a2 = ''
            l  = lower[idx]
            u  = upper[idx]
            k  = (a1,a2,l,u)
            L  = categories.get(k, [])
            L.append(idx)
            categories[k] = L
        ncategories = len(categories)
        # create data
        for idx, key in enumerate(categories):
            idxs = categories[key]
            data = self.data["bondsLength"][idxs]
            categories[key] = [str(d) for d in data]
        # adjust data size
        maxSize = max( [len(v) for v in categories.values()] )
        for key in categories:
            data = categories[key]
            add  = maxSize-len(data)
            if add > 0:
                categories[key] = data + ['']*add
        # start creating header and data
        sortCa = sorted( categories )
        header = [("%s%s%s(%.2f,%.2f)"%(a1,'-'*(len(a1)>0),a2,L,U)).replace(" ","_") for a1,a2,L,U in sortCa]
        data   = [categories[key] for key in sortCa]
        # save
        data = np.transpose(data)
        np.savetxt(fname     = fname,
                   X         = data,
                   fmt       = '%s',
                   delimiter = delimiter,
                   header    = " ".join(header),
                   comments  = comments)



# standard libraries imports
from __future__ import print_function
import os

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "engine.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    ENGINE.visualize(foldIntoBox=True, representationParams='VDW 0.1 20')
else:
    print(mes)



# standard libraries imports
from __future__ import print_function
import os

# import matplotlib
import matplotlib.pyplot as plt

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "engine.rmc")



# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    for frame in list(ENGINE.frames):
        ENGINE.set_used_frame(frame)
        GR, SQ, ACN, EMD = ENGINE.constraints
        GR.plot(ax=plt.figure().gca(), intra=False, show=False)
        SQ.plot(ax=plt.figure().gca(), intra=False, show=False)
        if ACN.used:
            ACN.plot(ax=plt.figure().gca(), show=False)
        if ACN.used:
            EMD.plot(ax=plt.figure().gca(), show=False)
    plt.show()
else:
    print(mes)



"""
This is a C compiled module to compute atomic bonds.
"""                      
from libc.math cimport sqrt, fabs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from fullrmc.Core.pairs_distances import pairs_distances_to_point, point_to_point_distance

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_ZERO      = 0.0
cdef C_FLOAT32 FLOAT_TWO       = 2.0
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_INT32   INT32_ZERO      = 0
cdef C_INT32   INT32_ONE       = 1




@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef C_FLOAT32 _single_bond(C_FLOAT32 bondLength,
                            C_FLOAT32 lower,
                            C_FLOAT32 upper,
                            bint      reduceDistanceToUpper = False,
                            bint      reduceDistanceToLower = False ):
    """
    It calculates the bonds constraint of a distances array.
    
    :Arguments:
       #. bondLength (float32): The bond distance
       #. lower (float32): The bond lower limit or minimum bond length allowed.
       #. upper (float32): The bond upper limit or maximum bond length allowed.
       #. reduceDistanceToUpper (bool): Whether to reduce bonds length found out of limits to the difference between the bond length and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceDistanceToLower (bool): Whether to reduce bonds length found out of limits to the difference between the bond length and the lower limit. When True, this flag may lose its priority for reduceDistanceToUpper if the later is True. DEFAULT: False
                  
    :Returns:
       #. reducedBond: The reduced bond length
    """
    # declare variables
    cdef C_FLOAT32 reducedBond
    # compute reducedBond
    if bondLength>=lower and bondLength<=upper:
        reducedBond = FLOAT_ZERO     
    elif reduceDistanceToUpper:
        reducedBond = fabs(upper-bondLength)
    elif reduceDistanceToLower:
        reducedBond = fabs(lower-bondLength)
    else:
        if bondLength > (lower+upper)/FLOAT_TWO:
            reducedBond = fabs(upper-bondLength)
        else:
            reducedBond = fabs(lower-bondLength)
    # return
    return reducedBond
    
 
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_bonds_coords( ndarray[C_INT32, ndim=1]   idx1 not None,
                       ndarray[C_INT32, ndim=1]   idx2 not None,
                       ndarray[C_FLOAT32, ndim=1] lowerLimit not None,
                       ndarray[C_FLOAT32, ndim=1] upperLimit not None,
                       ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                       ndarray[C_FLOAT32, ndim=2] basis not None,
                       bint                       isPBC,
                       bint                       reduceDistanceToUpper = False,
                       bint                       reduceDistanceToLower = False,
                       C_INT32                    ncores = 1):    
    """
    It calculates the bonds constraint of box coordinates.
    
    :Arguments:
       #. idx1 (int32 (n,) numpy.ndarray): First atoms index array
       #. idx2 (int32 (n,) numpy.ndarray): Second atoms index array
       #. lowerLimit (float32 (n,) numpy.ndarray): Lower limit or minimum bond length allowed.
       #. upperLimit (float32 (n,) numpy.ndarray): Upper limit or minimum bond length allowed.
       #. boxCoords (float32 (n,3) numpy.ndarray): The atomic coordinates array.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.       
       #. reduceDistanceToUpper (bool): Whether to reduce bonds length found out of limits to the difference between the bond length and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceDistanceToLower (bool): Whether to reduce bonds length found out of limits to the difference between the bond length and the lower limit. When True, this flag may lose its priority for reduceDistanceToUpper if the later is True. DEFAULT: False
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. bondsLength: The calculated bonds length
       #. reducedLengths: The calculated reduced bonds length
    """
    cdef C_INT32 i, numberOfIndexes
    cdef C_FLOAT32 bondLength, reducedLength
    numberOfIndexes = <C_INT32>len(lowerLimit)
    # create bondsLength and reducedBonds list
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] bondsLength    = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] reducedLengths = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32) 
    # loop atoms
    for i from 0 <= i < numberOfIndexes:
        bondLength  = point_to_point_distance( point1 = boxCoords[ idx1[i],: ], 
                                               point2 = boxCoords[ idx2[i],: ], 
                                               basis  = basis,
                                               isPBC  = isPBC)
        reducedLength = _single_bond(bondLength            = bondLength,
                                     lower                 = lowerLimit[i],
                                     upper                 = upperLimit[i],
                                     reduceDistanceToUpper = reduceDistanceToUpper,
                                     reduceDistanceToLower = reduceDistanceToLower )
        # append lists
        bondsLength[i]    = bondLength
        reducedLengths[i] = reducedLength
    # return bondsLength and reducedRistances
    return bondsLength, reducedLengths    
    
    
    









"""
This is a C compiled module to compute improper angles.
"""            
from libc.math cimport sqrt, fabs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from fullrmc.Core.pairs_distances import pair_difference_to_point, pairs_differences_to_point, from_to_points_differences

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_NEG_ONE   = -1.0
cdef C_FLOAT32 FLOAT_ZERO      = 0.0
cdef C_FLOAT32 FLOAT_ONE       = 1.0
cdef C_FLOAT32 FLOAT_TWO       = 2.0
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 PI              = 3.141592653589793
cdef C_FLOAT32 PI_2            = PI/2
cdef C_INT32   INT_ONE         = 1

 
cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x)
    C_FLOAT32 ceil(C_FLOAT32 x)
    C_FLOAT32 sqrt(C_FLOAT32 x)

cdef inline C_FLOAT32 round(C_FLOAT32 num):
    return floor(num + HALF_BOX_LENGTH) if (num > FLOAT_ZERO) else ceil(num - HALF_BOX_LENGTH)





@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef _single_improper_angle( ndarray[C_FLOAT32, ndim=1]    impVect , 
                             ndarray[C_FLOAT32, ndim=1]    oxVect , 
                             ndarray[C_FLOAT32, ndim=1]    oyVect , 
                             ndarray[C_FLOAT32, ndim=1]    lowerLimits,
                             ndarray[C_FLOAT32, ndim=1]    upperLimits,
                             ndarray[C_FLOAT32, ndim=1]    angles ,
                             ndarray[C_FLOAT32, ndim=1]    reducedAngles ,
                             C_INT32                       index,
                             bint                          reduceAngleToUpper = False,
                             bint                          reduceAngleToLower = False):

    # declare variables
    cdef C_FLOAT32 vectorNorm, dot
    cdef C_FLOAT32 angle, reducedAngle
    cdef C_FLOAT32 lower, upper
    cdef C_FLOAT32 improperVector_x, improperVector_y, improperVector_z
    cdef C_FLOAT32 oxVector_x, oxVector_y, oxVector_z
    cdef C_FLOAT32 oyVector_x, oyVector_y, oyVector_z
    cdef C_FLOAT32 ozVector_x, ozVector_y, ozVector_z
    ########################### normalize improper vector ###########################
    improperVector_x = impVect[0]
    improperVector_y = impVect[1]
    improperVector_z = impVect[2]
    vectorNorm = sqrt(improperVector_x*improperVector_x + improperVector_y*improperVector_y + improperVector_z*improperVector_z)
    if vectorNorm==0:
        raise Exception("Computing angle, improper vector found to have null length")
    improperVector_x /= vectorNorm
    improperVector_y /= vectorNorm
    improperVector_z /= vectorNorm
    ############################## normalize ox vector ##############################
    oxVector_x = oxVect[0]
    oxVector_y = oxVect[1]
    oxVector_z = oxVect[2]
    vectorNorm = sqrt(oxVector_x*oxVector_x + oxVector_y*oxVector_y + oxVector_z*oxVector_z)
    if vectorNorm==0:
        raise Exception("Computing angle, ox vector found to have null length")
    oxVector_x /= vectorNorm
    oxVector_y /= vectorNorm
    oxVector_z /= vectorNorm
    ############################## normalize oy vector ##############################
    oyVector_x = oyVect[0]
    oyVector_y = oyVect[1]
    oyVector_z = oyVect[2]
    vectorNorm = sqrt(oyVector_x*oyVector_x + oyVector_y*oyVector_y + oyVector_z*oyVector_z)
    if vectorNorm==0:
        raise Exception("Computing angle, oy vector found to have null length")
    oyVector_x /= vectorNorm
    oyVector_y /= vectorNorm
    oyVector_z /= vectorNorm
    ############################### compute oz vector ###############################
    # compute OZ vector as a×b= (a2b3−a3b2)i−(a1b3−a3b1)j+(a1b2−a2b1)k.
    ozVector_x =  oxVector_y*oyVector_z - oxVector_z*oyVector_y
    ozVector_y = -oxVector_x*oyVector_z + oxVector_z*oyVector_x
    ozVector_z =  oxVector_x*oyVector_y - oxVector_y*oyVector_x
    ################################ angle ################################
    # compute dot product
    dot = improperVector_x*ozVector_x + improperVector_y*ozVector_y + improperVector_z*ozVector_z
    # calculate angle
    angle = PI_2 - <C_FLOAT32>np.arccos( np.clip( dot ,-1, 1 ) ) # PI_2 - <C_FLOAT32>np.arccos( dot ) clipped for floating errors
    # compute reduced angle
    lower = lowerLimits[index]
    upper = upperLimits[index]
    if angle>=lower and angle<=upper:
        reducedAngle = FLOAT_ZERO     
    elif reduceAngleToUpper:
        reducedAngle = fabs(upper-angle)
    elif reduceAngleToLower:
        reducedAngle = fabs(lower-angle)
    else:
        if angle > (lower+upper)/FLOAT_TWO:
            reducedAngle = fabs(upper-angle)
        else:
            reducedAngle = fabs(lower-angle)
    # set angles and reduced
    angles[index]        = angle
    reducedAngles[index] = reducedAngle


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_improper_angles_coords( ndarray[C_INT32, ndim=1]   improperIdxs not None, 
                                 ndarray[C_INT32, ndim=1]   oIdxs not None,
                                 ndarray[C_INT32, ndim=1]   xIdxs not None,
                                 ndarray[C_INT32, ndim=1]   yIdxs not None,
                                 ndarray[C_FLOAT32, ndim=1] lowerLimit not None,
                                 ndarray[C_FLOAT32, ndim=1] upperLimit not None,
                                 ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                 ndarray[C_FLOAT32, ndim=2] basis not None,
                                 bint                       isPBC,
                                 bint                       reduceAngleToUpper = False,
                                 bint                       reduceAngleToLower = False,
                                 C_INT32                    ncores = 1):    
    """
    Computes the improper angles constraint between an improper atom and a plane atoms.
    The plane normal vector is calculated using the right-hand rule where (thumb=ox vector), 
    (index=oy vector) hence (oz=normal=second finger)
    
    :Arguments:
       #. improperIdxs (int32 (n,) numpy.ndarray): The improper atom indexes.
       #. oIdxs (int32 (n,) numpy.ndarray): The O atom indexes.
       #. xIdxs (int32 (n,) numpy.ndarray): The x atom indexes.
       #. yIdxs (int32 (n,) numpy.ndarray): The y atom indexes.
       #. lowerLimit (float32 (n,) numpy.ndarray): The angles lower limits.
       #. upperLimit (float32 (n,) numpy.ndarray): The angles upper limits.
       #. boxCoords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. reduceAngleToUpper (bool): Whether to reduce angle found out of limits to the difference between the angle and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceAngleToLower (bool): Whether to reduce angle found out of limits to the difference between the angle and the lower limit. When True, this flag may lose its priority for reduceAngleToUpper if the later is True. DEFAULT: False
       #. ncores (int32) [default=1]: The number of cores to use.
       
    :Returns:
       #. angles: The calculated angles (rad).
       #. reducedAngles: The reduced angles (rad)
    """
    
    cdef C_INT32 i, numberOfIndexes
    cdef C_FLOAT32 angle, reducedAngle
    # get number of indexes
    numberOfIndexes = <C_INT32>len(lowerLimit)
    # create abgles and reduced list
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] angles  = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] reduced = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32) 

    # loop all anglesf
    for i from 0 <= i < numberOfIndexes:
        impVect = pair_difference_to_point( point1 = boxCoords[improperIdxs[i],:], 
                                            point2 = boxCoords[oIdxs[i],:], 
                                            basis  = basis,
                                            isPBC  = isPBC,
                                            ncores = INT_ONE)  
        oxVect = pair_difference_to_point( point1 = boxCoords[oIdxs[i],:], 
                                           point2 = boxCoords[xIdxs[i],:], 
                                           basis  = basis,
                                           isPBC  = isPBC,
                                           ncores = INT_ONE)  
        oyVect = pair_difference_to_point( point1 = boxCoords[oIdxs[i],:], 
                                           point2 = boxCoords[yIdxs[i],:], 
                                           basis  = basis,
                                           isPBC  = isPBC,
                                           ncores = INT_ONE) 
        _single_improper_angle( impVect            = impVect ,
                                oxVect             = oxVect ,
                                oyVect             = oyVect ,
                                lowerLimits        = lowerLimit ,
                                upperLimits        = upperLimit ,
                                angles             = angles,
                                reducedAngles      = reduced,
                                index              = i ,
                                reduceAngleToUpper = reduceAngleToUpper,
                                reduceAngleToLower = reduceAngleToLower)  
    # return results
    return angles, reduced      

        



    
    
    
    
    
    
    
    
    
    
    
    

   
    


import os
from pdbparser.pdbparser import pdbparser
from pdbparser.Utilities.Construct import AmorphousSystem
from pdbparser.Utilities.Collection import get_path


# create thf amorphous box
pdb = pdbparser(os.path.join(get_path("pdbparser"),"Data/Tetrahydrofuran.pdb" ))
#pdb.visualize()
#exit()
pdb = AmorphousSystem(pdb, boxSize=[48,48,48], 
                           recursionLimit = 1000000,
                           insertionNumber=700,
                           density = 0.7,
                           priorities={"boxSize":True, "insertionNumber":False, "density":True}).construct().get_pdb()
pdb.export_pdb("thf.pdb")



"""
Group contains parent classes for all groups.
A Group is a set of atoms indexes used to gather atoms and apply actions
such as moves upon them. Therefore it has become possible to fully
customize and separate atoms to groups and perform stochastic actions
on groups rather than on single atoms.

.. inheritance-diagram:: fullrmc.Core.Group
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function
import inspect

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, get_path
from ..Core.MoveGenerator import MoveGenerator, RemoveGenerator
from ..Generators.Translations import TranslationGenerator
from ..Generators.Removes import AtomsRemoveGenerator

class Group(object):
    """
    A Group is a set of atoms indexes container.

    :Parameters:
        #. indexes (np.ndarray, list, set, tuple): list of atoms indexes.
        #. moveGenerator (None, MoveGenerator): Move generator instance.
           If None is given AtomsRemoveGenerator is considered by default.
        #. refine (bool): The refinement flag used by the Engine.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.Group import Group

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...

        # re-define groups as atoms.
        groups = [Group([idx]) for idx in ENGINE.pdb.indexes]
        ENGINE.set_groups( groups )

        # Re-define groups generators as needed ... By default AtomsRemoveGenerator is used.

    """
    def __init__(self, indexes, moveGenerator=None, refine=False):
        #self.__moveGenerator = TranslationGenerator(group=self)
        self.set_indexes(indexes)
        self.set_move_generator(moveGenerator)
        # set refine
        self.set_refine(refine)
        # initialize engine
        self.__engine = None

    def __len__(self):
        return len(self.__indexes)

    def _set_engine(self, engine):
        """ This is a fullrmc API method to set the group engine.
        Use with caution and better not to."""
        self.__engine = engine

    def _get_engine(self):
        """ This is a fullrmc API method to get the group engine.
        Normally a group doesn't need to know the engine."""
        return self.__engine

    @property
    def indexes(self):
        """ Atoms index array."""
        return self.__indexes

    @property
    def moveGenerator(self):
        """ Group's move generator instance."""
        return self.__moveGenerator

    @property
    def refine(self):
        """ Refine flag."""
        return self.__refine

    def set_refine(self, refine):
        """
        Set the selector refine flag.

        :Parameters:
            #. refine (bool): The selector refinement flag.
        """
        assert isinstance(refine, bool), LOGGER.error("refine must be a boolean")
        self.__refine = refine

    def set_indexes(self, indexes):
        """
        Set group atoms index. Indexes redundancy is not checked
        and indexes order is preserved.

        :Parameters:
            #. indexes (list,set,tuple,np.ndarray): The group atoms indexes.
        """
        assert isinstance(indexes, (list,set,tuple,np.ndarray)), LOGGER.error("indexes must be either a list or a numpy.array")
        if isinstance(indexes, np.ndarray):
            assert len(indexes.shape) == 1, LOGGER.error("each group must be a numpy.ndarray of dimension 1")
        # create group of indexes
        group = []
        for idx in list(indexes):
            assert is_integer(idx), LOGGER.error("group indexes must be integers")
            assert idx>=0, LOGGER.error("group index must equal or bigger than 0")
            group.append(int(idx))
        # check length
        assert len(group) or isinstance(self, EmptyGroup), LOGGER.error("Group is found to be empty. Use EmptyGroup instead.")
        # create indexes
        self.__indexes = np.array(group, dtype=INT_TYPE)

    def set_move_generator(self, generator):
        """
        Set group move generator.

        :Parameters:
            #. generator (None, MoveGenerator): Move generator instance.
               If None is given TranslationGenerator is considered by default.
        """
        if generator is None:
            generator = TranslationGenerator(group=self)
        else:
            assert isinstance(generator, MoveGenerator), LOGGER.error("generator must be a MoveGenerator instance")
            assert not isinstance(generator, RemoveGenerator), LOGGER.error("generator must not be a RemoveGenerator instance")
            try:
                generator.set_group(self)
            except Exception as e:
                raise Exception( LOGGER.error(e) )
        # set moveGenerator
        self.__moveGenerator = generator


class EmptyGroup(Group):
    """
    Empty group is a special group that takes no atoms indexes. It's mainly
    used to remove atoms from system upon fitting.

    :Parameters:
        #. indexes (object): list of atoms indexes. This argument will always
           be disregarded in this particular case.
        #. moveGenerator (None, MoveGenerator): Move generator instance.
           If None is given RemoveGenerator is considered by default.
        #. refine (bool): The refinement flag used by the Engine.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.Group import EmptyGroup

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...

        # re-define groups and set a single empty group
        ENGINE.set_groups( EmptyGroup() )

        # Re-define groups generators as needed ... By default RemoveGenerator is used.

    """
    def __init__(self, indexes=None, moveGenerator=None, refine=False):
        super(EmptyGroup, self).__init__(indexes=None,
                                         moveGenerator=moveGenerator,
                                         refine=refine)

    @property
    def moveGenerator(self):
        """ Group's move generator instance."""
        return self.__moveGenerator

    @property
    def indexes(self):
        """ Always returns None for EmptyGroup"""
        return self.__indexes

    def set_move_generator(self, generator):
        """
        Set group move generator.

        :Parameters:
            #. generator (None, MoveGenerator): Move generator instance.
               If None is given TranslationGenerator is considered by default.
        """
        if generator is None:
            generator = AtomsRemoveGenerator(group=self)
        else:
            assert isinstance(generator, RemoveGenerator), LOGGER.error("EmptyGroup generator must be a RemoveGenerator instance")
        self.__moveGenerator = generator

    def set_indexes(self, indexes):
        """
        Sets the group indexes. For an EmptyGroup, this method will disregard
        given indexes argument and will always set indexes property to None.

        :Parameters:
            #. indexes (object): The group atoms indexes. This argument will always be
               disregarded in this particular case.
        """
        self.__indexes = None



"""
This is a C compiled module to compute atomic pair distances.
"""
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from cython.parallel import prange

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 BOX_LENGTH      = 1.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 FLOAT32_ZERO    = 0.0
cdef C_INT32   INT32_ZERO      = 0
cdef C_INT32   INT32_ONE       = 1


cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x) nogil
    C_FLOAT32 ceil(C_FLOAT32 x)  nogil
    C_FLOAT32 sqrt(C_FLOAT32 x)  nogil

    
cdef inline C_FLOAT32 round(C_FLOAT32 num) nogil:
    return floor(num + HALF_BOX_LENGTH) if (num > FLOAT32_ZERO) else ceil(num - HALF_BOX_LENGTH)


      
############################################################################################        
################################# C DIFFERENCE DEFINITIONS #################################

######## From To differences ########
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _from_to_boxpoint_realdifference_PBC( C_FLOAT32[:]   boxPointFrom,
                                                C_FLOAT32[:]   boxPointTo,
                                                C_FLOAT32[:]   difference,
                                                C_FLOAT32[:,:] basis):
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    # calculate difference
    diff_x = boxPointTo[0]-boxPointFrom[0]
    diff_y = boxPointTo[1]-boxPointFrom[1]
    diff_z = boxPointTo[2]-boxPointFrom[2]
    box_dx = diff_x-round(diff_x)
    box_dy = diff_y-round(diff_y)
    box_dz = diff_z-round(diff_z)
    # get real difference
    difference[0] = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
    difference[1] = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
    difference[2] = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]




@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _from_to_boxpoints_realdifferences_PBC( C_FLOAT32[:,:] boxPointsFrom,
                                                  C_FLOAT32[:,:] boxPointsTo,
                                                  C_FLOAT32[:,:] basis,
                                                  C_FLOAT32[:,:] differences,
                                                  C_INT32        ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_INT32 num_threads = ncores
    # loop
    for i in prange(<C_INT32>boxPointsFrom.shape[0], nogil=True, schedule="static", num_threads=num_threads): 
    #for i from INT32_ZERO <= i < <C_INT32>boxPointsFrom.shape[0]:
        # calculate difference
        diff_x = boxPointsTo[i,0]-boxPointsFrom[i,0]
        diff_y = boxPointsTo[i,1]-boxPointsFrom[i,1]
        diff_z = boxPointsTo[i,2]-boxPointsFrom[i,2]
        box_dx = diff_x-round(diff_x)
        box_dy = diff_y-round(diff_y)
        box_dz = diff_z-round(diff_z)
        # get real difference
        differences[i,0] = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        differences[i,1] = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        differences[i,2] = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _from_to_realpoint_realdifference_PBC( C_FLOAT32[:] realPointFrom,
                                                 C_FLOAT32[:] realPointTo,
                                                 C_FLOAT32[:] difference):
    # calculate difference
    difference[0] = realPointTo[0]-realPointFrom[0]
    difference[1] = realPointTo[1]-realPointFrom[1]
    difference[2] = realPointTo[2]-realPointFrom[2]
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _from_to_realpoints_realdifferences_IBC( C_FLOAT32[:,:] realPointsFrom,
                                                   C_FLOAT32[:,:] realPointsTo,
                                                   C_FLOAT32[:,:] differences,
                                                   C_INT32        ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_INT32 num_threads = ncores
    # loop
    for i in prange(INT32_ZERO, <C_INT32>realPointsFrom.shape[0], INT32_ONE, nogil=True, schedule="static", num_threads=num_threads):
    #for i from INT32_ZERO <= i < <C_INT32>boxPointsFrom.shape[0]:
        # calculate difference
        differences[i,0] = realPointsTo[i,0]-realPointsFrom[i,0]
        differences[i,1] = realPointsTo[i,1]-realPointsFrom[i,1]
        differences[i,2] = realPointsTo[i,2]-realPointsFrom[i,2]
        
        
        
######## point differences ########       
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _boxcoords_realdifferences_to_boxpoint_PBC( C_FLOAT32[:]   boxPoint,
                                                      C_FLOAT32[:,:] boxCoords,
                                                      C_FLOAT32[:,:] basis,
                                                      C_FLOAT32[:,:] differences,
                                                      C_INT32        ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_INT32 num_threads = ncores
    # loop
    for i in prange(<C_INT32>boxCoords.shape[0], nogil=True, schedule="static", num_threads=num_threads): 
    #for i in range(<C_INT32>boxCoords.shape[0]):
        # calculate difference
        diff_x = boxPoint[0]-boxCoords[i,0]
        diff_y = boxPoint[1]-boxCoords[i,1]
        diff_z = boxPoint[2]-boxCoords[i,2]
        box_dx = diff_x-round(diff_x)
        box_dy = diff_y-round(diff_y)
        box_dz = diff_z-round(diff_z)
        # get real difference
        differences[i,0] = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        differences[i,1] = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        differences[i,2] = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _realcoords_realdifferences_to_realpoint_IBC( C_FLOAT32[:]   realPoint,
                                                        C_FLOAT32[:,:] realCoords,
                                                        C_FLOAT32[:,:] differences,
                                                        C_INT32        ncores=1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 atomReal_x, atomReal_y, atomReal_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomReal_x = realPoint[0]
    atomReal_y = realPoint[1]
    atomReal_z = realPoint[2]    
    # loop
    for i in prange(<C_INT32>realCoords.shape[0], nogil=True, schedule="static", num_threads=num_threads): 
    #for i in range(<C_INT32>realCoords.shape[0]):
        # calculate difference
        differences[i,0] = atomReal_x-realCoords[i,0]
        differences[i,1] = atomReal_y-realCoords[i,1]
        differences[i,2] = atomReal_z-realCoords[i,2]
        
        
            
######## index differences ########
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)               
cdef void _boxcoords_realdifferences_to_indexcoords_PBC( C_INT32        atomIndex,
                                                         C_FLOAT32[:,:] boxCoords,
                                                         C_FLOAT32[:,:] basis,
                                                         C_FLOAT32[:,:] differences,
                                                         bint           allAtoms = True,
                                                         C_INT32        ncores = 1) nogil:                          
    # declare variables
    cdef C_INT32 i, startIndex, endIndex
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_FLOAT32 atomBox_x, atomBox_y, atomBox_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomBox_x = boxCoords[atomIndex,0]
    atomBox_y = boxCoords[atomIndex,1]
    atomBox_z = boxCoords[atomIndex,2] 
    # start index
    if allAtoms:
        startIndex = INT32_ZERO
    else:
        startIndex = <C_INT32>atomIndex
    endIndex = <C_INT32>boxCoords.shape[0]
    # loop
    for i in prange(startIndex, endIndex, INT32_ONE, nogil=True, schedule="static", num_threads=num_threads): 
    #for i from startIndex <= i < endIndex:
        # calculate difference
        diff_x = atomBox_x-boxCoords[i,0]
        diff_y = atomBox_y-boxCoords[i,1]
        diff_z = atomBox_z-boxCoords[i,2]
        box_dx = diff_x-round(diff_x)
        box_dy = diff_y-round(diff_y)
        box_dz = diff_z-round(diff_z)
        # get real difference
        differences[i,0] = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        differences[i,1] = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        differences[i,2] = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]                                                  
               
                                                    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _realcoords_realdifferences_to_indexcoords_IBC( C_INT32        atomIndex,
                                                          C_FLOAT32[:,:] realCoords,
                                                          C_FLOAT32[:,:] differences,
                                                          bint           allAtoms = True,
                                                          C_INT32        ncores=1) nogil:   
    # declare variables
    cdef C_INT32 i, startIndex, endIndex
    cdef C_FLOAT32 atomReal_x, atomReal_y, atomReal_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomReal_x = realCoords[atomIndex,0]
    atomReal_y = realCoords[atomIndex,1]
    atomReal_z = realCoords[atomIndex,2] 
    # start index
    if allAtoms:
        startIndex = INT32_ZERO
    else:
        startIndex = <C_INT32>atomIndex
    endIndex = <C_INT32>realCoords.shape[0]
    # loop
    for i in prange(startIndex, endIndex, INT32_ONE, nogil=True, schedule="static", num_threads=num_threads): 
    #for i from startIndex <= i < endIndex:
        # calculate difference
        differences[i,0] = atomReal_x - realCoords[i,0]
        differences[i,1] = atomReal_y - realCoords[i,1]
        differences[i,2] = atomReal_z - realCoords[i,2]

   

   
############################################################################################        
################################## C DISTANCE DEFINITIONS ##################################
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef C_FLOAT32 _boxcoords_realdistance_to_boxpoint_PBC( C_FLOAT32[:]   boxPoint1,
                                                        C_FLOAT32[:]   boxPoint2,
                                                        C_FLOAT32[:,:] basis) nogil:
    # declare variables
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_FLOAT32 real_dx, real_dy, real_dz,
    # compute difference    
    diff_x = boxPoint1[0] - boxPoint2[0]
    diff_y = boxPoint1[1] - boxPoint2[1]
    diff_z = boxPoint1[2] - boxPoint2[2]
    box_dx = diff_x - round(diff_x)
    box_dy = diff_y - round(diff_y)
    box_dz = diff_z - round(diff_z)
    # get real difference
    real_dx = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
    real_dy = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
    real_dz = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]
    # compute and return return distance
    return <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)

     
                
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _boxcoords_realdistances_to_boxpoint_PBC( C_FLOAT32[:]   boxPoint,
                                                    C_FLOAT32[:,:] boxCoords,
                                                    C_FLOAT32[:,:] basis,
                                                    C_FLOAT32[:]   distances,
                                                    C_INT32        ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_FLOAT32 real_dx, real_dy, real_dz,
    cdef C_FLOAT32 atomBox_x, atomBox_y, atomBox_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomBox_x = boxPoint[0]
    atomBox_y = boxPoint[1]
    atomBox_z = boxPoint[2]    
    # loop
    for i in prange(<C_INT32>boxCoords.shape[0], nogil=True, schedule="static", num_threads=num_threads): # Error compiling Cython file: Cannot read reduction variable in loop body
    #for i in range(<C_INT32>boxCoords.shape[0]):
        # calculate difference
        diff_x = atomBox_x - boxCoords[i,0]
        diff_y = atomBox_y - boxCoords[i,1]
        diff_z = atomBox_z - boxCoords[i,2]
        box_dx = diff_x - round(diff_x)
        box_dy = diff_y - round(diff_y)
        box_dz = diff_z - round(diff_z)
        # get real difference
        real_dx = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        real_dy = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        real_dz = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]
        # calculate distance         
        distances[i] = <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)               
cdef void _boxcoords_realdistances_to_indexcoords_PBC( C_INT32        atomIndex,
                                                       C_FLOAT32[:,:] boxCoords,
                                                       C_FLOAT32[:,:] basis,
                                                       C_FLOAT32[:]   distances,
                                                       bint           allAtoms = True,
                                                       C_INT32        ncores = 1) nogil:                          
    # declare variables
    cdef C_INT32 i, startIndex, endIndex
    cdef C_FLOAT32 diff_x, diff_y, diff_z
    cdef C_FLOAT32 box_dx, box_dy, box_dz
    cdef C_FLOAT32 real_dx, real_dy, real_dz,
    cdef C_FLOAT32 atomBox_x, atomBox_y, atomBox_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomBox_x = boxCoords[atomIndex,0]
    atomBox_y = boxCoords[atomIndex,1]
    atomBox_z = boxCoords[atomIndex,2] 
    # start index
    if allAtoms:
        startIndex = INT32_ZERO
    else:
        startIndex = <C_INT32>atomIndex
    endIndex = <C_INT32>boxCoords.shape[0]
    # loop
    for i in prange(startIndex, endIndex, INT32_ONE, nogil=True, schedule="static", num_threads=num_threads): # Error compiling Cython file: Cannot read reduction variable in loop body
    #for i from startIndex <= i < endIndex:
        # calculate difference
        diff_x = atomBox_x - boxCoords[i,0]
        diff_y = atomBox_y - boxCoords[i,1]
        diff_z = atomBox_z - boxCoords[i,2]
        box_dx = diff_x - round(diff_x)
        box_dy = diff_y - round(diff_y)
        box_dz = diff_z - round(diff_z)
        # get real difference
        real_dx = box_dx*basis[0,0] + box_dy*basis[1,0] + box_dz*basis[2,0]
        real_dy = box_dx*basis[0,1] + box_dy*basis[1,1] + box_dz*basis[2,1]
        real_dz = box_dx*basis[0,2] + box_dy*basis[1,2] + box_dz*basis[2,2]
        # calculate distance         
        distances[i] = <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)                                                    
                                                    



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef C_FLOAT32 _realcoords_realdistance_to_realpoint_IBC( C_FLOAT32[:]   realPoint1,
                                                          C_FLOAT32[:]   realPoint2) nogil:
    # declare variables
    cdef C_FLOAT32 real_dx, real_dy, real_dz
    # loop
    real_dx = realPoint1[0]-realPoint2[0]
    real_dy = realPoint1[1]-realPoint2[1]
    real_dz = realPoint1[2]-realPoint2[2]
    # calculate and return distance         
    return <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)


                                     
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _realcoords_realdistances_to_realpoint_IBC( C_FLOAT32[:]   realPoint,
                                                      C_FLOAT32[:,:] realCoords,
                                                      C_FLOAT32[:]   distances,
                                                      C_INT32        ncores=1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 real_dx, real_dy, real_dz
    cdef C_FLOAT32 atomReal_x, atomReal_y, atomReal_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomReal_x = realPoint[0]
    atomReal_y = realPoint[1]
    atomReal_z = realPoint[2]    
    # loop
    for i in prange(<C_INT32>realCoords.shape[0], nogil=True, schedule="static", num_threads=num_threads): 
    #for i in range(<C_INT32>realCoords.shape[0]):
        # calculate difference
        real_dx = realCoords[i,0]-atomReal_x
        real_dy = realCoords[i,1]-atomReal_y
        real_dz = realCoords[i,2]-atomReal_z
        # calculate distance         
        distances[i] = <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)

        
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef void _realcoords_realdistances_to_indexcoords_IBC( C_INT32        atomIndex,
                                                        C_FLOAT32[:,:] realCoords,
                                                        C_FLOAT32[:]   distances,
                                                        bint           allAtoms = True,
                                                        C_INT32        ncores=1) nogil:   
    # declare variables
    cdef C_INT32 i, startIndex, endIndex
    cdef C_FLOAT32 real_dx, real_dy, real_dz
    cdef C_FLOAT32 atomReal_x, atomReal_y, atomReal_z
    cdef C_INT32 num_threads = ncores
    # get point coordinates
    atomReal_x = realCoords[atomIndex,0]
    atomReal_y = realCoords[atomIndex,1]
    atomReal_z = realCoords[atomIndex,2] 
    # start index
    if allAtoms:
        startIndex = INT32_ZERO
    else:
        startIndex = <C_INT32>atomIndex
    endIndex = <C_INT32>realCoords.shape[0]
    # loop
    for i in prange(startIndex, endIndex, INT32_ONE, nogil=True, schedule="static", num_threads=num_threads): 
    #for i from startIndex <= i < endIndex:
        # calculate difference
        real_dx = atomReal_x - realCoords[i,0]
        real_dy = atomReal_y - realCoords[i,1]
        real_dz = atomReal_z - realCoords[i,2]
        # calculate distance         
        distances[i] = <C_FLOAT32>sqrt(real_dx*real_dx + real_dy*real_dy + real_dz*real_dz)
        
        
        
        
############################################################################################        
############################## PYTHON DIFFERENCE DEFINITIONS ###############################
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def from_to_points_differences( ndarray[C_FLOAT32, ndim=2] pointsFrom not None,
                                ndarray[C_FLOAT32, ndim=2] pointsTo not None,
                                C_FLOAT32[:,:]             basis not None,
                                bint                       isPBC, 
                                C_INT32                    ncores = 1):                       
    """
    Compute point to point vector difference between two atomic coordinates arrays taking 
    into account periodic or infinite boundary conditions. Difference is calculated as 
    the following:
    
    .. math::
            differences[i,:] = boundaryConditions( pointsTo[i,:] - pointsFrom[i,:] )
    
    :Arguments:
       #. pointsFrom (float32 (n,3) numpy.ndarray): The first atomic coordinates array of 
          the same shape as pointsTo.
       #. pointsTo (float32 (n,3) numpy.ndarray): The second atomic coordinates array of 
          the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3) numpy.ndarray): The computed differences array.   
    """
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] differences = np.empty((<C_INT32>pointsFrom.shape[0],3), dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        _from_to_boxpoints_realdifferences_PBC( boxPointsFrom = pointsFrom,
                                                boxPointsTo   = pointsTo,
                                                basis         = basis,
                                                differences   = differences,
                                                ncores        = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _from_to_realpoints_realdifferences_IBC( realPointsFrom = pointsFrom,
                                                 realPointsTo   = pointsTo,
                                                 differences    = differences,
                                                 ncores         = ncores)
    # return differences
    return differences



# 
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pair_difference_to_point( ndarray[C_FLOAT32, ndim=1] point1  not None,
                              ndarray[C_FLOAT32, ndim=1] point2 not None,
                              C_FLOAT32[:,:]             basis not None,
                              bint                       isPBC, 
                              C_INT32                    ncores = 1):                        
    """
    Compute differences between one atomic coordinates arrays to a point coordinates  
    taking into account periodic or infinite boundary conditions. Difference is 
    calculated as the following:
    
    .. math::
            differences[i,:] = boundaryConditions( point[0,:] - coords[i,:] )
    
    :Arguments:
       #. point (float32 (1,3) numpy.ndarray): The atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3) numpy.ndarray): The computed differences array.   
    """
    # if periodic boundary conditions, coords must be in box
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] difference = np.empty((3), dtype=NUMPY_FLOAT32)
    if isPBC:
        _from_to_boxpoint_realdifference_PBC( boxPointFrom = point1,
                                              boxPointTo   = point2,
                                              difference   = difference,
                                              basis        = basis)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _from_to_realpoint_realdifference_PBC( realPointFrom  = point1,
                                               realPointTo    = point2,
                                               difference     = difference)
    # return difference
    return difference



 
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_differences_to_point( ndarray[C_FLOAT32, ndim=1] point  not None,
                                ndarray[C_FLOAT32, ndim=2] coords not None,
                                C_FLOAT32[:,:]             basis not None,
                                bint                       isPBC, 
                                C_INT32                    ncores = 1):                        
    """
    Compute differences between one atomic coordinates arrays to a point coordinates  
    taking into account periodic or infinite boundary conditions. Difference is 
    calculated as the following:
    
    .. math::
            differences[i,:] = boundaryConditions( point[0,:] - coords[i,:] )
    
    :Arguments:
       #. point (float32 (1,3) numpy.ndarray): The atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3) numpy.ndarray): The computed differences array.   
    """
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] differences = np.empty((<C_INT32>coords.shape[0],3), dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        _boxcoords_realdifferences_to_boxpoint_PBC( boxPoint    = point,
                                                    boxCoords   = coords,
                                                    basis       = basis,
                                                    differences = differences,
                                                    ncores      = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _realcoords_realdifferences_to_realpoint_IBC( realPoint   = point,
                                                      realCoords  = coords,
                                                      differences = differences,
                                                      ncores      = ncores)
    # return differences
    return differences


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_differences_to_indexcoords( C_INT32                    atomIndex,
                                      ndarray[C_FLOAT32, ndim=2] coords not None,
                                      C_FLOAT32[:,:]             basis not None,
                                      bint                       isPBC,
                                      bint                       allAtoms = True,
                                      C_INT32                    ncores = 1):                       
    """
    Compute differences between one atomic coordinates arrays to a point coordinates   
    given its index in the coordinates array and taking into account periodic or 
    infinite boundary conditions. Difference is calculated as the following:
    
    .. math::
            differences[i,:] = boundaryConditions( coords[atomIndex,:] - coords[i,:] )
    
    :Arguments:
       #. atomIndex (int32): The index of the atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3) numpy.ndarray): The computed differences array.   
    """
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] differences = np.empty((<C_INT32>coords.shape[0],3), dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        _boxcoords_realdifferences_to_indexcoords_PBC( atomIndex   = atomIndex,
                                                       boxCoords   = coords,
                                                       basis       = basis,
                                                       differences = differences,
                                                       allAtoms    = allAtoms,
                                                       ncores      = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _realcoords_realdifferences_to_indexcoords_IBC( atomIndex   = atomIndex,
                                                        realCoords  = coords,
                                                        differences = differences,
                                                        allAtoms    = allAtoms,
                                                        ncores      = ncores)
    # return differences
    return differences    

    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_differences_to_multi_points( ndarray[C_FLOAT32, ndim=2] points not None,
                                       ndarray[C_FLOAT32, ndim=2] coords not None,
                                       C_FLOAT32[:,:]             basis not None,
                                       bint                       isPBC, 
                                       C_INT32                    ncores = 1):                       
    """
    Compute differences between one atomic coordinates arrays to a multiple points   
    coordinates taking into account periodic or infinite boundary conditions. 
    Difference is calculated as the following:
    
    .. math::
            differences[i,:,k] = boundaryConditions( points[k,:] - coords[i,:] )
    
    :Arguments:
       #. points (float32 (k,3) numpy.ndarray): The multiple atomic coordinates points.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3,k) numpy.ndarray): The computed differences array.   
    """
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] differences = np.empty( (<C_INT32>coords.shape[0], 
                                                                        3,
                                                                        <C_INT32>points.shape[1]), 
                                                                        dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        for i from INT32_ZERO <= i < <C_INT32>points.shape[1]:
            _boxcoords_realdifferences_to_boxpoint_PBC( boxPoint  = points[:,i],
                                                      boxCoords   = coords,
                                                      basis       = basis,
                                                      differences = differences[:,:,i],
                                                      ncores      = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        for i from INT32_ZERO <= i < <C_INT32>points.shape[1]:
            _realcoords_realdifferences_to_realpoint_IBC( realPoint   = points[:,i],
                                                          realCoords  = coords,
                                                          differences = differences[:,:,i],
                                                          ncores      = ncores)
    # return differences
    return differences    
    
    
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_differences_to_multi_indexcoords( ndarray[C_INT32, ndim=1]   indexes,
                                            ndarray[C_FLOAT32, ndim=2] coords not None,
                                            C_FLOAT32[:,:]             basis not None,
                                            bint                       isPBC,
                                            bint                       allAtoms = True,
                                            C_INT32                    ncores = 1):                       
    """
    Compute differences between one atomic coordinates arrays to a points coordinates   
    given their indexes in the coordinates array and taking into account periodic or 
    infinite boundary conditions. Difference is calculated as the following:
    
    .. math::
            differences[i,:,k] = boundaryConditions( coords[indexes[k],:] - coords[i,:] )
    
    :Arguments:
       #. indexes (int32 (k,3) numpy.ndarray): The atomic coordinates indexes array.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. differences (float32 (n,3,k) numpy.ndarray): The computed differences array.   
    """
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=3] differences = np.empty( (<C_INT32>coords.shape[0], 
                                                                        3,
                                                                        <C_INT32>indexes.shape[0]), 
                                                                        dtype=NUMPY_FLOAT32)                                                                      
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        for i from INT32_ZERO <= i < <C_INT32>indexes.shape[0]:
            _boxcoords_realdifferences_to_indexcoords_PBC( atomIndex   = indexes[i],
                                                           boxCoords   = coords,
                                                           basis       = basis,
                                                           differences = differences[:,:,i],
                                                           allAtoms    = allAtoms,
                                                           ncores      = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        for i from INT32_ZERO <= i < <C_INT32>indexes.shape[0]:
            _realcoords_realdifferences_to_indexcoords_IBC( atomIndex   = indexes[i],
                                                            realCoords  = coords,
                                                            differences = differences[:,:,i],
                                                            allAtoms    = allAtoms,
                                                            ncores      = ncores)
    # return differences
    return differences 
    
    

    
############################################################################################        
############################### PYTHON DISTANCE DEFINITIONS ################################
@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def point_to_point_distance( ndarray[C_FLOAT32, ndim=1] point1  not None,
                             ndarray[C_FLOAT32, ndim=1] point2 not None,
                             C_FLOAT32[:,:]             basis not None,
                             bint                       isPBC, 
                             C_INT32                    ncores = 1):                       
    
    """
    Compute distances between one atomic coordinates arrays to a point coordinates  
    taking into account periodic or infinite boundary conditions. Distances is 
    calculated as the following:
    
    .. math::
            distances[i] = \sqrt{ \sum_{d}^{3}{ boundaryConditions( point[0,d] - coords[i,d] )^{2}} }
    
    :Arguments:
       #. point (float32 (1,3) numpy.ndarray): The atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. distances (float32 (n,) numpy.ndarray): The computed distances array.   
    """
    # if periodic boundary conditions, coords must be in box
    if isPBC:
       return _boxcoords_realdistance_to_boxpoint_PBC( boxPoint1 = point1,
                                                       boxPoint2 = point2,
                                                       basis     = basis)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        return _realcoords_realdistance_to_realpoint_IBC( realPoint1  = point1,
                                                          realPoint2  = point2)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_distances_to_point( ndarray[C_FLOAT32, ndim=1] point  not None,
                              ndarray[C_FLOAT32, ndim=2] coords not None,
                              C_FLOAT32[:,:]             basis not None,
                              bint                       isPBC, 
                              C_INT32                    ncores = 1):                       
    
    """
    Compute distances between one atomic coordinates arrays to a point coordinates  
    taking into account periodic or infinite boundary conditions. Distances is 
    calculated as the following:
    
    .. math::
            distances[i] = \sqrt{ \sum_{d}^{3}{ boundaryConditions( point[0,d] - coords[i,d] )^{2}} }
    
    :Arguments:
       #. point (float32 (1,3) numpy.ndarray): The atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. distances (float32 (n,) numpy.ndarray): The computed distances array.   
    """
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] distances = np.empty((<C_INT32>coords.shape[0],), dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        _boxcoords_realdistances_to_boxpoint_PBC( boxPoint  = point,
                                                  boxCoords = coords,
                                                  basis     = basis,
                                                  distances = distances,
                                                  ncores    = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _realcoords_realdistances_to_realpoint_IBC( realPoint  = point,
                                                    realCoords = coords,
                                                    distances  = distances,
                                                    ncores     = ncores)
    # return distances
    return distances


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_distances_to_indexcoords( C_INT32                    atomIndex,
                                    ndarray[C_FLOAT32, ndim=2] coords not None,
                                    C_FLOAT32[:,:]             basis not None,
                                    bint                       isPBC,
                                    bint                       allAtoms = True,
                                    C_INT32                    ncores = 1):                       
    
    """
    Compute distances between one atomic coordinates arrays to a points coordinates   
    given their indexes in the coordinates array and taking into account periodic or 
    infinite boundary conditions. Distances is calculated as the following:
    
    .. math::
            distances[i] = \sqrt{ \sum_{d}^{3}{ boundaryConditions( coords[atomIndex[i],d] - coords[i,d] )^{2}} }
    
    :Arguments:
       #. point (float32 (1,3) numpy.ndarray): The atomic coordinates point.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. distances (float32 (n,) numpy.ndarray): The computed distances array.   
    """
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] distances = np.empty((<C_INT32>coords.shape[0],), dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        _boxcoords_realdistances_to_indexcoords_PBC( atomIndex = atomIndex,
                                                     boxCoords = coords,
                                                     basis     = basis,
                                                     distances = distances,
                                                     allAtoms  = allAtoms,
                                                     ncores    = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        _realcoords_realdistances_to_indexcoords_IBC( atomIndex  = atomIndex,
                                                      realCoords = coords,
                                                      distances  = distances,
                                                      allAtoms   = allAtoms,
                                                      ncores     = ncores)
    # return distances
    return distances    


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_distances_to_multi_points( ndarray[C_FLOAT32, ndim=2] points not None,
                                     ndarray[C_FLOAT32, ndim=2] coords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC, 
                                     C_INT32                    ncores = 1):                       
    """
    Compute distances between one atomic coordinates arrays to a multiple points   
    coordinates taking into account periodic or infinite boundary conditions. 
    Distances is calculated as the following:
    
    .. math::
            distances[i,k] = \sqrt{ \sum_{d}^{3}{ boundaryConditions( points[k,d] - coords[i,d] )^{2}} }
            
    :Arguments:
       #. points (float32 (k,3) numpy.ndarray): The multiple atomic coordinates points.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. distances (float32 (n,) numpy.ndarray): The computed distances array.   
    """
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] distances = np.empty( (<C_INT32>coords.shape[0], 
                                                                      <C_INT32>points.shape[1]), 
                                                                      dtype=NUMPY_FLOAT32)
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        for i from INT32_ZERO <= i < <C_INT32>points.shape[1]:
            _boxcoords_realdistances_to_boxpoint_PBC( boxPoint  = points[:,i],
                                                      boxCoords = coords,
                                                      basis     = basis,
                                                      distances = distances[:,i],
                                                      ncores    = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        for i from INT32_ZERO <= i < <C_INT32>points.shape[1]:
            _realcoords_realdistances_to_realpoint_IBC( realPoint  = points[:,i],
                                                        realCoords = coords,
                                                        distances  = distances[:,i],
                                                        ncores     = ncores)
    # return distances
    return distances    
    

@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def pairs_distances_to_multi_indexcoords( ndarray[C_INT32, ndim=1]   indexes,
                                          ndarray[C_FLOAT32, ndim=2] coords not None,
                                          C_FLOAT32[:,:]             basis not None,
                                          bint                       isPBC,
                                          bint                       allAtoms = True,
                                          C_INT32                    ncores = 1):                       
    """
    Compute distances between one atomic coordinates arrays to a points coordinates   
    given their indexes in the coordinates array and taking into account periodic or 
    infinite boundary conditions. Distances is calculated as the following:
    
    .. math::
            distances[i,k] = \sqrt{ \sum_{d}^{3}{ boundaryConditions( coords[indexes[k],:]  - coords[i,d] )^{2}} }
            
    :Arguments:
       #. indexes (int32 (k,3) numpy.ndarray): The atomic coordinates indexes array.
       #. coords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. ncores (int32) [default=1]: The number of cores to use. 
       
    :Returns:
       #. distances (float32 (n,) numpy.ndarray): The computed distances array.   
    """
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=2] distances = np.empty( (<C_INT32>coords.shape[0], 
                                                                      <C_INT32>indexes.shape[0]), 
                                                                      dtype=NUMPY_FLOAT32)                                                                      
    # if periodic boundary conditions, coords must be in box
    if isPBC:
        for i from INT32_ZERO <= i < <C_INT32>indexes.shape[0]:
            _boxcoords_realdistances_to_indexcoords_PBC( atomIndex = indexes[i],
                                                         boxCoords = coords,
                                                         basis     = basis,
                                                         distances = distances[:,i],
                                                         allAtoms  = allAtoms,
                                                         ncores    = ncores)
    # if infinite boundary conditions coords must be in Cartesian normal space
    else:
        for i from INT32_ZERO <= i < <C_INT32>indexes.shape[0]:
            _realcoords_realdistances_to_indexcoords_IBC( atomIndex  = indexes[i],
                                                          realCoords = coords,
                                                          distances  = distances[:,i],
                                                          allAtoms   = allAtoms,
                                                          ncores     = ncores)
    # return distances
    return distances 



    


"""
This is a C compiled module to compute improper angles.
"""            
from libc.math cimport sqrt, fabs
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from numpy cimport ndarray
from fullrmc.Core.pairs_distances import  pair_difference_to_point

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_ZERO      = 0.0
cdef C_FLOAT32 FLOAT_TWO       = 2.0
cdef C_FLOAT32 HALF_BOX_LENGTH = 0.5
cdef C_FLOAT32 PI              = 3.141592653589793
cdef C_FLOAT32 PI_DEG          = 180.
cdef C_FLOAT32 FULL_CIRCLE     = 360.
cdef C_FLOAT32 RAD_TO_DEG      = PI_DEG/PI
cdef C_INT32   INT_ONE         = 1

 
cdef extern from "math.h":
    C_FLOAT32 floor(C_FLOAT32 x)
    C_FLOAT32 ceil(C_FLOAT32 x)
    C_FLOAT32 sqrt(C_FLOAT32 x)

cdef inline C_FLOAT32 round(C_FLOAT32 num):
    return floor(num + HALF_BOX_LENGTH) if (num > FLOAT_ZERO) else ceil(num - HALF_BOX_LENGTH)





@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef _single_dihedral_angle( ndarray[C_FLOAT32, ndim=1]    b1 , 
                             ndarray[C_FLOAT32, ndim=1]    b2 , 
                             ndarray[C_FLOAT32, ndim=1]    b3 , 
                             C_FLOAT32                     lowerLimit1,
                             C_FLOAT32                     upperLimit1,
                             C_FLOAT32                     lowerLimit2,
                             C_FLOAT32                     upperLimit2,
                             C_FLOAT32                     lowerLimit3,
                             C_FLOAT32                     upperLimit3,
                             ndarray[C_FLOAT32, ndim=1]    angles ,
                             ndarray[C_FLOAT32, ndim=1]    reducedAngles ,
                             C_INT32                       index,
                             bint                          reduceAngleToUpper = False,
                             bint                          reduceAngleToLower = False):

    # declare variables
    cdef C_FLOAT32 n1_x, n1_y, n1_z
    cdef C_FLOAT32 n2_x, n2_y, n2_z
    cdef C_FLOAT32 m1_x, m1_y, m1_z
    cdef C_FLOAT32 var, norm, x, y, angle, reducedAngle    
    # normalize b1
    norm = sqrt(b1[0]*b1[0] + b1[1]*b1[1] + b1[2]*b1[2])
    if norm==0:
        raise Exception("dihedral b1 vector is found to be zero length")
    b1[0] /= norm
    b1[1] /= norm
    b1[2] /= norm
    # normalize b2
    norm = sqrt(b2[0]*b2[0] + b2[1]*b2[1] + b2[2]*b2[2])
    if norm==0:
        raise Exception("dihedral b2 vector is found to be zero length")
    b2[0] /= norm
    b2[1] /= norm
    b2[2] /= norm
    # normalize b3
    norm = sqrt(b3[0]*b3[0] + b3[1]*b3[1] + b3[2]*b3[2])
    if norm==0:
        raise Exception("dihedral b3 vector is found to be zero length")
    b3[0] /= norm
    b3[1] /= norm
    b3[2] /= norm
    # compute n1 (b1 X b2)
    n1_x = b1[1] * b2[2] - b1[2] * b2[1]
    n1_y = b1[2] * b2[0] - b1[0] * b2[2]
    n1_z = b1[0] * b2[1] - b1[1] * b2[0]
    # compute n2 (b2 X b3)
    n2_x = b2[1] * b3[2] - b2[2] * b3[1]
    n2_y = b2[2] * b3[0] - b2[0] * b3[2]
    n2_z = b2[0] * b3[1] - b2[1] * b3[0]
    # compute m1 (n1 X b2) so n1,m1 and b2 form an orthogonal  frame
    m1_x = n1_y * b2[2] - n1_z * b2[1]
    m1_y = n1_z * b2[0] - n1_x * b2[2]
    m1_z = n1_x * b2[1] - n1_y * b2[0]
    # compute n2 in (n1,m1,b2) frame as x=n1.n2 and y=m1.n2 and z should be 0
    x = n1_x*n2_x + n1_y*n2_y + n1_z*n2_z
    y = m1_x*n2_x + m1_y*n2_y + m1_z*n2_z
    # compute dihedral angle as atan2(y,x) and convert to degrees
    angle = <C_FLOAT32>np.arctan2( y,x )* RAD_TO_DEG
    # convert to between 0-360
    if angle<0:
        angle = FULL_CIRCLE+angle
    # compute reduced angle trying first shell
    if (upperLimit1<lowerLimit1) and angle>upperLimit1:
        reducedAngle = FLOAT_ZERO
    elif angle>=lowerLimit1 and angle<=upperLimit1:
        reducedAngle = FLOAT_ZERO
    else:
        if reduceAngleToUpper:
            reducedAngle = <C_FLOAT32>fabs(upperLimit1-angle)
        elif reduceAngleToLower:
            reducedAngle = <C_FLOAT32>fabs(lowerLimit1-angle)
        else:
            if angle > (lowerLimit1+upperLimit1)/FLOAT_TWO:
                reducedAngle = <C_FLOAT32>fabs(upperLimit1-angle)
            else:
                reducedAngle = <C_FLOAT32>fabs(lowerLimit1-angle)
    # compute reduced angle trying second shell
    if reducedAngle!=FLOAT_ZERO:
        if (upperLimit2<lowerLimit2) and angle>upperLimit2:
            reducedAngle = FLOAT_ZERO
        elif angle>=lowerLimit2 and angle<=upperLimit2:
            reducedAngle = FLOAT_ZERO
        else:
            if reduceAngleToUpper:
                var = <C_FLOAT32>fabs(upperLimit2-angle)
            elif reduceAngleToLower:
                var = <C_FLOAT32>fabs(lowerLimit2-angle)
            else:
                if angle > (lowerLimit2+upperLimit2)/FLOAT_TWO:
                    var = <C_FLOAT32>fabs(upperLimit2-angle)
                else:
                    var = <C_FLOAT32>fabs(lowerLimit2-angle)
            if var<reducedAngle:
                reducedAngle = var
    # compute reduced angle trying third and last shell
    if reducedAngle!=FLOAT_ZERO:
        if (upperLimit3<lowerLimit3) and angle>upperLimit3:
            reducedAngle = FLOAT_ZERO
        elif angle>=lowerLimit3 and angle<=upperLimit3:
            reducedAngle = FLOAT_ZERO
        else:
            if reduceAngleToUpper:
                var = <C_FLOAT32>fabs(upperLimit3-angle)
            elif reduceAngleToLower:
                var = <C_FLOAT32>fabs(lowerLimit3-angle)
            else:
                if angle > (lowerLimit3+upperLimit3)/FLOAT_TWO:
                    var = <C_FLOAT32>fabs(upperLimit3-angle)
                else:
                    var = <C_FLOAT32>fabs(lowerLimit3-angle)
            if var<reducedAngle:
                reducedAngle = var
    # set angles and reduced
    angles[index]        = angle
    reducedAngles[index] = reducedAngle


@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def full_dihedral_angles_coords( ndarray[C_INT32, ndim=1]   indexes1 not None, 
                                 ndarray[C_INT32, ndim=1]   indexes2 not None,
                                 ndarray[C_INT32, ndim=1]   indexes3 not None,
                                 ndarray[C_INT32, ndim=1]   indexes4 not None,
                                 ndarray[C_FLOAT32, ndim=1] lowerLimit1 not None,
                                 ndarray[C_FLOAT32, ndim=1] upperLimit1 not None,
                                 ndarray[C_FLOAT32, ndim=1] lowerLimit2 not None,
                                 ndarray[C_FLOAT32, ndim=1] upperLimit2 not None,
                                 ndarray[C_FLOAT32, ndim=1] lowerLimit3 not None,
                                 ndarray[C_FLOAT32, ndim=1] upperLimit3 not None,
                                 ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                 ndarray[C_FLOAT32, ndim=2] basis not None,
                                 bint                       isPBC,
                                 bint                       reduceAngleToUpper = False,
                                 bint                       reduceAngleToLower = False,
                                 C_INT32                    ncores = 1):    
    """
    Computes the improper angles constraint between an improper atom and a plane atoms.
    The plane normal vector is calculated using the right-hand rule where (thumb=ox vector), 
    (index=oy vector) hence (oz=normal=second finger)
    
    :Arguments:
       #. indexes1 (int32 (n,) numpy.ndarray): Diherdral first atom indexes.
       #. indexes2 (int32 (n,) numpy.ndarray): Diherdral second atom indexes.
       #. indexes3 (int32 (n,) numpy.ndarray): Diherdral third atom indexes.
       #. indexes4 (int32 (n,) numpy.ndarray): Diherdral fourth atom indexes.
       #. lowerLimit1 (float32 (n,) numpy.ndarray): First shells lower limit.
       #. upperLimit1 (float32 (n,) numpy.ndarray): First shells upper limits.
       #. lowerLimit2 (float32 (n,) numpy.ndarray): Second shells lower limit.
       #. upperLimit2 (float32 (n,) numpy.ndarray): Second shells upper limits.
       #. lowerLimit3 (float32 (n,) numpy.ndarray): Third shells lower limit.
       #. upperLimit3 (float32 (n,) numpy.ndarray): Third shells upper limits.
       #. boxCoords (float32 (n,3) numpy.ndarray): The atomic coordinates array of the same shape as pointsFrom.
       #. basis (float32 (3,3) numpy.ndarray): The (3x3) boundary conditions box vectors.
       #. isPBC (bool): Whether it is a periodic boundary conditions or infinite.
       #. reduceAngleToUpper (bool): Whether to reduce angle found out of limits to the difference between the angle and the upper limit. When True, this flag has the higher priority. DEFAULT: False
       #. reduceAngleToLower (bool): Whether to reduce angle found out of limits to the difference between the angle and the lower limit. When True, this flag may lose its priority for reduceAngleToUpper if the later is True. DEFAULT: False
       #. ncores (int32) [default=1]: The number of cores to use.
       
    :Returns:
       #. angles: The calculated angles (rad).
       #. reducedAngles: The reduced angles (rad)
    """
    
    cdef C_INT32 i, numberOfIndexes
    cdef C_FLOAT32 angle, reducedAngle
    # get number of indexes
    numberOfIndexes = <C_INT32>indexes1.shape[0]
    # create abgles and reduced list
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] angles  = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32)
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] reduced = np.zeros((numberOfIndexes), dtype=NUMPY_FLOAT32) 

    # loop all angles
    for i from 0 <= i < numberOfIndexes:   
        b1 = pair_difference_to_point( point1 = boxCoords[indexes1[i],:], 
                                       point2 = boxCoords[indexes2[i],:], 
                                       basis  = basis,
                                       isPBC  = isPBC,
                                       ncores = INT_ONE) 
        b2 = pair_difference_to_point( point1 = boxCoords[indexes2[i],:], 
                                       point2 = boxCoords[indexes3[i],:], 
                                       basis  = basis,
                                       isPBC  = isPBC,
                                       ncores = INT_ONE) 
        b3 = pair_difference_to_point( point1 = boxCoords[indexes3[i],:], 
                                       point2 = boxCoords[indexes4[i],:], 
                                       basis  = basis,
                                       isPBC  = isPBC,
                                       ncores = INT_ONE)                                                                                                                      
        _single_dihedral_angle( b1                 = b1 ,
                                b2                 = b2 ,
                                b3                 = b3 ,
                                lowerLimit1        = lowerLimit1[i] ,
                                upperLimit1        = upperLimit1[i] ,
                                lowerLimit2        = lowerLimit2[i] ,
                                upperLimit2        = upperLimit2[i] ,
                                lowerLimit3        = lowerLimit3[i] ,
                                upperLimit3        = upperLimit3[i] ,
                                angles             = angles,
                                reducedAngles      = reduced,
                                index              = i ,
                                reduceAngleToUpper = reduceAngleToUpper,
                                reduceAngleToLower = reduceAngleToLower)  
    # return results
    return angles, reduced      

        



    
    
    
    
    
    
    
    
    
    
    
    

   
    


##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# imports
import os, sys
import numpy as np
from fullrmc.Engine import Engine
from fullrmc.Constraints.PairDistributionConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.AtomicCoordinationConstraints import AtomicCoordinationNumberConstraint
from fullrmc.Selectors.RandomSelectors import RandomSelector
from fullrmc.Core.GroupSelector import RecursiveGroupSelector
from fullrmc.Selectors.OrderedSelectors import DirectionalOrderSelector



##########################################################################################
#####################################  CREATE ENGINE  ####################################
# dirname
try:
    DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
except:
    DIR_PATH = ''

# files name
grFileName     = "SiOx.gr"
pdbFileName    = "SiOx.pdb"
engineFileName = "SiOx.rmc"
NCORES         = 1
FRESH_START    = False
# engine variables
grExpPath      = os.path.join(DIR_PATH, grFileName)
pdbPath        = os.path.join(DIR_PATH, pdbFileName)
engineFilePath = os.path.join(DIR_PATH, engineFileName)


ENGINE = Engine(path=None)
if not ENGINE.is_engine(engineFilePath) or FRESH_START:
   # create engine
    ENGINE = Engine(path=engineFilePath, freshStart=True)
    ENGINE.set_pdb(pdbPath)
    # create and add pair distribution constraint to the engine
    PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=grExpPath, weighting="atomicNumber")
    # shape function parameters
    params = {'rmin':0., 'rmax':None, 'dr':0.5,
              'qmin':0.0001, 'qmax':0.6, 'dq':0.005,
              'updateFreq':1000}
    PDF_CONSTRAINT.set_shape_function_parameters(params)
    ENGINE.add_constraints([PDF_CONSTRAINT])
    # Intermolecular constraint
    EMD_CONSTRAINT = InterMolecularDistanceConstraint()
    ENGINE.add_constraints([EMD_CONSTRAINT])
    EMD_CONSTRAINT.set_type_definition("element")
    EMD_CONSTRAINT.set_pairs_distance([('Si','Si',1.75), ('o','o',1.10), ('Si','o',1.30)])
    # coordination number constraint
    ACNC_CONSTRAINT = AtomicCoordinationNumberConstraint()
    ENGINE.add_constraints([ACNC_CONSTRAINT])
    ACNC_CONSTRAINT.set_coordination_number_definition( coordNumDef=[('Si','Si',1.8,2.6,3,5), ])
    # initialize constraints
    ENGINE.initialize_used_constraints()
    # set number density
    ENGINE.set_number_density(.0125)
    # save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(engineFilePath)
    # get constraints
    PDF_CONSTRAINT, EMD_CONSTRAINT, ACNC_CONSTRAINT = ENGINE.constraints


ACNC_CONSTRAINT.set_used(False)
##########################################################################################
#####################################  DIFFERENT RUNS  ###################################
def run_normal_rmc(nsteps=150000, saveFreq=10000, ncores=1):
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=saveFreq, ncores=ncores, restartPdb=None)

def run_explore(recur=50, saveFreq=10000, ncores=1):
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=False, explore=True)
    ENGINE.set_group_selector(gs)
     # number of steps
    nsteps = recur*len(ENGINE.groups)
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, ncores=ncores, restartPdb=None)

def expand_nanoparticule(recur=50, explore=True, refine=False, ncores=1):
    # create expansion selector and explore creating shell layer
    center = np.sum(ENGINE.realCoordinates, axis=0)/ENGINE.realCoordinates.shape[0]
    GS  = DirectionalOrderSelector(ENGINE, center=center, expand=True, adjustMoveGenerators=True)
    RGS = RecursiveGroupSelector(GS, recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(RGS)
    nsteps = ENGINE.numberOfAtoms*recur
    # run rmc
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps/10, ncores=ncores, restartPdb=None)



##########################################################################################
#####################################  RUN SIMULATION  ###################################
run_normal_rmc(nsteps=100000, ncores=NCORES)
run_explore(ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
run_explore(ncores=NCORES)
# expand nanoparticule
expand_nanoparticule(ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
expand_nanoparticule(ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
expand_nanoparticule(ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
expand_nanoparticule(ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)
run_normal_rmc(nsteps=100000, ncores=NCORES)


##########################################################################################
###################################### CALL plot.py ######################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



# standard libraries imports
from __future__ import print_function
import os

# fullrmc library imports
from fullrmc import Engine

# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "engine.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if result:
    ENGINE = ENGINE.load(engineFilePath)
    GR, SQ, CN, MD = ENGINE.constraints
    CN.compute_data()
    GR.plot(intra=False,show=False)
    SQ.plot(intra=False,show=False)
    CN.plot(show=False)
    MD.plot(show=True)
else:
    print(mes)



# standard libraries imports
from __future__ import print_function
import os
import itertools

# external libraries imports
import wx
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

# fullrmc library imports
from fullrmc.Engine import Engine
from fullrmc.Constraints.PairCorrelationConstraints import PairDistributionConstraint


# Create plotting styles
#styles  = ['-','--','-.',':']
colors = ["b",'g','r','c','m','y']
markers = ["",'.','+','^','|']
INTRA_STYLES = [r[0] + r[1]for r in itertools.product(['--'], colors)]
INTRA_STYLES = [r[0] + r[1]for r in itertools.product(markers, INTRA_STYLES)]
INTER_STYLES = [r[0] + r[1]for r in itertools.product(['-'], colors)]
INTER_STYLES = [r[0] + r[1]for r in itertools.product(markers, INTER_STYLES)]


# dirname
DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
engineFilePath = os.path.join(DIR_PATH, "CO2.rmc")

# load
ENGINE = Engine(path=None)
result, mes = ENGINE.is_engine(engineFilePath, mes=True)
if not result:
    print(mes)
    exit()

# load engine and get pdf constraint engine
ENGINE = ENGINE.load(engineFilePath)
PDF_CONSTRAINT = ENGINE.constraints[0]

def create_figure(PDF):
    # get output
    output = PDF.get_constraint_value()
    # create figure
    FIG = plt.figure()
    FIG.patch.set_facecolor('white')
    grid = gridspec.GridSpec(nrows=2, ncols=2)
    grid.update(left=0.05, right=0.95, wspace=0.05)
    totalAx = plt.subplot(grid[0, :])
    lowRAx  = plt.subplot(grid[1, 0])
    highRAx = plt.subplot(grid[1, 1])
    # set axis ticks
    highRAx.get_yaxis().set_ticks([])
    lowRAx.get_yaxis().set_ticks([])
    # plot experimental
    totalAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
    lowRAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
    highRAx.plot(PDF.experimentalDistances, ENGINE.constraints[0].experimentalPDF, 'ro', label="observed", markersize=7.5, markevery=1 )
    # plot partials
    intraStyleIndex = 0
    interStyleIndex = 0
    for key in output:
        val = output[key]
        if key in ("pdf_total", "pdf"):
            continue
        elif "intra" in key:
            totalAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            lowRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            highRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            intraStyleIndex+=1
        elif "inter" in key:
            totalAx.plot(PDF.shellCenters, val, INTER_STYLES[interStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            lowRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            highRAx.plot(PDF.shellCenters, val, INTRA_STYLES[intraStyleIndex], markevery=5, label=key.split("rdf_")[1] )
            interStyleIndex+=1
    # plot totals
    totalAx.plot(PDF.shellCenters, output["pdf"], 'k', linewidth=3.0,  markevery=25, label="total" )
    lowRAx.plot(PDF.shellCenters, output["pdf"], 'k', linewidth=3.0,  markevery=25, label="total" )
    highRAx.plot(PDF.shellCenters, output["pdf"], 'k', linewidth=3.0,  markevery=25, label="total" )
    # set legend
    totalAx.legend(ncol=2, frameon=False, fontsize=12)#, loc=(1.05,-1.25))
    # remove y ticks labels
    lowRAx.set_yticklabels([])
    highRAx.set_yticklabels([])
    # set x limits
    lowRAx.relim()
    highRAx.relim()
    lowRAx.set_xlim([0,1.4])
    highRAx.set_xlim([1.4,5])
    totalAx.set_xlim([0,10])
    # set y limits
    lowRAx.set_ylim(top=4.9)
    highRAx.set_ylim(top=1.5)
    # set title
    if ENGINE.totalStandardError is not None:  plt.suptitle("$StdErr=%.6f$"%ENGINE.totalStandardError, fontsize=20)
    return FIG, totalAx, lowRAx, highRAx

def add_low_r_ax_annotations(ax):
    # annotate lowRAx
    ax.annotate('', xy=(0.67, 1.1), xytext=(1.35, -0.45), ha="right",
                      arrowprops=dict(facecolor='black',
                                      arrowstyle="simple",
                                      fc="w", ec="k",
                                      connectionstyle="arc3,rad=1.7") )
    ax.text(0.45, 3.6,'explore allows tunneling\n and going through energy\n barriers from a peak \nto another.',
                horizontalalignment='center',
                verticalalignment='center')

def add_total_ax_annotations(ax):
    ax.annotate('', xy=(0.725, 1.1), xytext=(3.5, 2), ha="right",
                         arrowprops=dict(facecolor='black',
                                         arrowstyle="-|>",
                                         fc="w", ec="k",
                                         connectionstyle="arc3,rad=-0.2") )

    ax.annotate('', xy=(1.1, 4), xytext=(3.5, 3.5), ha="right",
                         arrowprops=dict(facecolor='black',
                                         arrowstyle="-|>",
                                         fc="w", ec="k",
                                         connectionstyle="arc3,rad=0.2") )

    ax.text(3.5, 2.75,'Bonding electron density\n polarizability measured by X-rays',
                 horizontalalignment='center',
                 verticalalignment='center')

# get output
FIG, totalAx, lowRAx, highRAx = create_figure(PDF_CONSTRAINT)
add_low_r_ax_annotations(lowRAx)
add_total_ax_annotations(totalAx)


# show
plt.show()
